import struct
import time
from abc import ABC, abstractmethod
from functools import wraps
from typing import TYPE_CHECKING, Callable, Dict, List, Optional, Tuple, Union

import ingenialogger
import numpy as np
from ingenialink.enums.register import REG_DTYPE

from ingeniamotion.enums import (
    MonitoringProcessStage,
    MonitoringSoCConfig,
    MonitoringSoCType,
    MonitoringVersion,
)
from ingeniamotion.exceptions import IMMonitoringError
from ingeniamotion.metaclass import DEFAULT_AXIS, DEFAULT_SERVO

if TYPE_CHECKING:
    from ingeniamotion.motion_controller import MotionController


def check_monitoring_disabled(func: Callable[..., None]) -> Callable[..., None]:
    @wraps(func)
    def wrapper(self, *args, **kwargs):  # type: ignore
        monitoring_enabled = self.mc.capture.is_monitoring_enabled(servo=self.servo)
        if monitoring_enabled:
            raise IMMonitoringError("Monitoring is enabled")
        return func(self, *args, **kwargs)

    return wrapper


class Monitoring(ABC):
    """Class to configure a monitoring in a servo.

    Args:
        mc (MotionController): MotionController instance.
        servo (str): servo alias to reference it. ``default`` by default.

    """

    REGISTER_MAP_OFFSET = 0x800
    ESTIMATED_MAX_TIME_FOR_SAMPLE = 0.003

    _data_type_size = {
        REG_DTYPE.U8: 1,
        REG_DTYPE.S8: 1,
        REG_DTYPE.U16: 2,
        REG_DTYPE.S16: 2,
        REG_DTYPE.U32: 4,
        REG_DTYPE.S32: 4,
        REG_DTYPE.U64: 8,
        REG_DTYPE.S64: 8,
        REG_DTYPE.FLOAT: 4,
    }

    MONITORING_FREQUENCY_DIVIDER_REGISTER = "MON_DIST_FREQ_DIV"
    MONITORING_NUMBER_MAPPED_REGISTERS_REGISTER = "MON_CFG_TOTAL_MAP"
    MONITOR_START_CONDITION_TYPE_REGISTER = "MON_CFG_SOC_TYPE"
    MONITORING_INDEX_CHECKER_REGISTER = "MON_IDX_CHECK"
    MONITORING_TRIGGER_DELAY_SAMPLES_REGISTER = "MON_CFG_TRIGGER_DELAY"
    MONITORING_WINDOW_NUMBER_SAMPLES_REGISTER = "MON_CFG_WINDOW_SAMP"
    MONITORING_ACTUAL_NUMBER_SAMPLES_REGISTER = "MON_CFG_CYCLES_VALUE"
    MONITORING_FORCE_TRIGGER_REGISTER = "MON_CMD_FORCE_TRIGGER"

    def __init__(self, mc: "MotionController", servo: str = DEFAULT_SERVO) -> None:
        super().__init__()
        self.mc = mc
        self.servo = servo
        self.mapped_registers: List[Dict[str, Union[int, str, REG_DTYPE]]] = []
        self.sampling_freq: Optional[float] = None
        self._read_process_finished = False
        self.samples_number = 0
        self.trigger_delay_samples = 0
        self.logger = ingenialogger.get_logger(__name__, drive=mc.servo_name(servo))
        self.max_sample_number = mc.capture.monitoring_max_sample_size(servo)
        self.data = None
        self._version: Optional[MonitoringVersion] = None

    @check_monitoring_disabled
    def set_frequency(self, prescaler: int) -> None:
        """Function to define monitoring frequency with a prescaler. Frequency will be
        ``Position & velocity loop rate frequency / prescaler``, see
        :func:`ingeniamotion.configuration.Configuration.get_position_and_velocity_loop_rate`
        to know about this frequency. Monitoring must be disabled.

        Args:
            prescaler : determines monitoring frequency.
                It must be ``1`` or higher.

        Raises:
            ValueError: If prescaler is less than ``1``.

        """
        if prescaler < 1:
            raise ValueError("prescaler must be 1 or higher")
        position_velocity_loop_rate = self.mc.configuration.get_position_and_velocity_loop_rate(
            servo=self.servo, axis=DEFAULT_AXIS
        )
        self.sampling_freq = round(position_velocity_loop_rate / prescaler, 2)
        self.mc.communication.set_register(
            self.MONITORING_FREQUENCY_DIVIDER_REGISTER, prescaler, servo=self.servo, axis=0
        )

    @check_monitoring_disabled
    def map_registers(self, registers: List[Dict[str, Union[int, str, REG_DTYPE]]]) -> None:
        """Map registers to monitoring. Monitoring must be disabled.

        Args:
            registers : List of registers to map.
                Each register must be a dict with two keys:

                .. code-block:: python

                    {
                        "name": "CL_POS_FBK_VALUE",  # Register name.
                        "axis": 1  # Register axis.
                        # If it has no axis field, by default axis 1.
                    }

        Raises:
            IMMonitoringError: If register maps fails in the servo.
            IMMonitoringError: If buffer size is not enough for all the registers.
            TypeError: If some parameter has a wrong type.

        """
        drive = self.mc.servos[self.servo]
        drive.monitoring_remove_all_mapped_registers()

        for channel in registers:
            subnode = channel.get("axis", DEFAULT_AXIS)
            if not isinstance(subnode, int):
                raise TypeError("Subnode has to be an integer")
            register = channel["name"]
            if not isinstance(register, str):
                raise TypeError("Register has to be a string")
            register_obj = self.mc.info.register_info(register, subnode, servo=self.servo)
            dtype = register_obj.dtype
            channel["dtype"] = dtype

        self._check_buffer_size_is_enough(
            self.samples_number, self.trigger_delay_samples, registers
        )

        for ch_idx, channel in enumerate(registers):
            subnode = channel.get("axis", DEFAULT_AXIS)
            if not isinstance(subnode, int):
                raise TypeError("Subnode has to be an integer")
            register = channel["name"]
            if not isinstance(register, str):
                raise TypeError("Register has to be a string")
            if not isinstance(channel["dtype"], REG_DTYPE):
                raise TypeError("dtype has to be of type REG_DTYPE")
            dtype = channel["dtype"]
            register_obj = self.mc.info.register_info(register, subnode, servo=self.servo)
            drive.monitoring_set_mapped_register(
                ch_idx,
                register_obj.mapped_address,
                subnode,
                dtype.value,
                self._data_type_size[dtype],
            )

        num_mon_reg = self.mc.communication.get_register(
            self.MONITORING_NUMBER_MAPPED_REGISTERS_REGISTER, servo=self.servo, axis=0
        )
        if not isinstance(num_mon_reg, int):
            raise TypeError("Number of mapped registers value has to be an integer")
        if num_mon_reg < 1:
            raise IMMonitoringError("Map Monitoring registers fails")
        self.mapped_registers = registers

    @abstractmethod
    @check_monitoring_disabled
    def set_trigger(
        self,
        trigger_mode: MonitoringSoCType,
        edge_condition: Optional[MonitoringSoCConfig] = None,
        trigger_signal: Optional[Dict[str, Union[int, str]]] = None,
        trigger_value: Union[None, int, float] = None,
    ) -> None:
        """Configure monitoring trigger. Monitoring must be disabled.

        Args:
            trigger_mode : monitoring start of condition type.
            edge_condition : edge event type. ``None`` by default.
            trigger_signal : dict with name and axis of trigger signal
                for rising or falling edge trigger. ``None`` by default.
            trigger_value : value for rising or falling edge trigger.
                ``None`` by default.

        Raises:
            TypeError: If trigger_mode is trigger event edge and
                edge_condition, trigger_signal or trigger_value are None.
            IMMonitoringError: If trigger signal is not mapped.

        """
        pass

    def _get_reg_index_and_edge_condition_value(
        self, trigger_signal: Dict[str, str], trigger_value: Union[int, float]
    ) -> Tuple[int, int]:
        index_reg = -1
        for index, item in enumerate(self.mapped_registers):
            if (
                trigger_signal["name"] == item["name"]
                and trigger_signal.get("axis", DEFAULT_AXIS) == item["axis"]
            ):
                index_reg = index
        if index_reg < 0:
            raise IMMonitoringError("Trigger signal is not mapped in Monitoring")
        dtype = self.mapped_registers[index_reg]["dtype"]
        if not isinstance(dtype, REG_DTYPE):
            raise TypeError("dtype has to be of type REG_DTYPE")
        level_edge = self._unpack_trigger_value(trigger_value, dtype)
        return index_reg, level_edge

    @staticmethod
    def _unpack_trigger_value(value: Union[int, float], dtype: REG_DTYPE) -> int:
        """Converts any value from its dtype to an UINT32"""
        if dtype == REG_DTYPE.U16:
            return int(np.array([int(value)], dtype="int64").astype("uint16")[0])
        if dtype == REG_DTYPE.U32:
            return int(struct.unpack("L", struct.pack("I", int(value)))[0])
        if dtype == REG_DTYPE.S32:
            if value < 0:
                return int(value + (1 << 32))
            return int(np.array([int(value)], dtype="int64").astype("int32")[0])
        return int(struct.unpack("L", struct.pack("f", value))[0])

    @abstractmethod
    @check_monitoring_disabled
    def configure_number_samples(self, total_num_samples: int, trigger_delay_samples: int) -> None:
        """Configure monitoring number of samples. Monitoring must be disabled.

        Args:
            total_num_samples : monitoring total number of samples.
            trigger_delay_samples : monitoring number of samples before trigger.
                It should be less than total_num_samples. Minimum ``0``.

        Raises:
            ValueError: If trigger_delay_samples is less than ``0``
                or higher than total_num_samples.
            IMMonitoringError: If buffer size is not enough for all the samples.

        """
        pass

    @check_monitoring_disabled
    def configure_sample_time(self, total_time: float, trigger_delay: float) -> None:
        """Configure monitoring number of samples defines by sample and trigger
        delay time. Monitoring must be disabled.

        Args:
            total_time : monitoring sample total time, in seconds.
            trigger_delay : trigger delay in seconds. Value should be
                between ``-total_time/2`` and  ``total_time/2``.

        Raises:
            ValueError: If trigger_delay is not between ``-total_time/2`` and
                ``total_time/2``.
            IMMonitoringError: If buffer size is not enough for all the samples.
            TypeError: If some parameter has a wrong type.

        """
        if total_time / 2 < abs(trigger_delay):
            raise ValueError("trigger_delay value should be between -total_time/2 and total_time/2")
        if not isinstance(self.sampling_freq, float):
            raise TypeError("Sampling frequency has to be set before configuring the sample time")
        total_num_samples = int(self.sampling_freq * total_time)
        trigger_delay_samples = int(((total_time / 2) - trigger_delay) * self.sampling_freq)
        self.configure_number_samples(total_num_samples, trigger_delay_samples)

    def _check_trigger_timeout(self, init_time: float, timeout: Optional[float]) -> None:
        if timeout is None:
            return
        time_now = time.time()
        if init_time + timeout < time_now:
            self.logger.warning("Timeout. No trigger was reached.")
            self._read_process_finished = True

    def _check_read_data_timeout(self, init_read_time: float) -> None:
        time_now = time.time()
        total_num_samples = len(self.mapped_registers) * self.samples_number
        max_timeout = self.ESTIMATED_MAX_TIME_FOR_SAMPLE * total_num_samples
        if init_read_time + max_timeout < time_now:
            self.logger.warning("Timeout. Drive take too much time reading data")
            self._read_process_finished = True

    def _check_read_data_ends(self, data_length: int) -> None:
        if data_length >= self.samples_number:
            self._read_process_finished = True

    def _update_read_process_finished(
        self,
        init_read_time: Optional[float],
        current_len: int,
        init_time: float,
        timeout: Optional[float],
    ) -> None:
        self._check_read_data_ends(current_len)
        if self._read_process_finished:
            return
        if init_read_time is None:
            self._check_trigger_timeout(init_time, timeout)
        else:
            self._check_read_data_timeout(init_read_time)

    def _show_current_process(
        self,
        current_len: int,
        progress_callback: Optional[Callable[[MonitoringProcessStage, float], None]],
    ) -> None:
        process_stage = self.mc.capture.get_monitoring_process_stage(
            servo=self.servo, version=self._version
        )
        current_progress = current_len / self.samples_number
        if process_stage in [
            MonitoringProcessStage.DATA_ACQUISITION,
            MonitoringProcessStage.END_STAGE,
        ]:
            self.logger.debug("Read %.2f%% of monitoring data", current_progress * 100)
        if progress_callback is not None:
            progress_callback(process_stage, current_progress)

    @abstractmethod
    def _check_monitoring_is_ready(self) -> Tuple[bool, Optional[str]]:
        pass

    @abstractmethod
    def _check_data_is_ready(self) -> bool:
        pass

    # TODO Study remove progress_callback
    def read_monitoring_data(
        self,
        timeout: Optional[float] = None,
        progress_callback: Optional[Callable[[MonitoringProcessStage, float], None]] = None,
    ) -> List[List[Union[int, float]]]:
        """Blocking function that read the monitoring data.

        Args:
            timeout : maximum time trigger is waited, in seconds.
                ``None`` by default.

        Returns:
            Data of monitoring. Each element of the list is a different register data.

        """
        drive = self.mc.servos[self.servo]
        self._read_process_finished = False
        is_ready, result_text = self._check_monitoring_is_ready()
        data_array: List[List[Union[int, float]]] = [[] for _ in self.mapped_registers]
        self.logger.debug("Waiting for data")
        init_read_time, init_time = None, time.time()
        current_len = 0
        while not self._read_process_finished:
            if self._check_data_is_ready():
                init_read_time = init_read_time or time.time()
                drive.monitoring_read_data()
                self._fill_data(data_array)
                current_len = len(data_array[0])
            elif not is_ready:
                self.logger.warning(result_text)
                self._read_process_finished = True
            self._update_read_process_finished(init_read_time, current_len, init_time, timeout)
            self._show_current_process(current_len, progress_callback)
        return data_array

    def _fill_data(self, data_array: List[List[Union[int, float]]]) -> None:
        drive = self.mc.servos[self.servo]
        for ch_idx, channel in enumerate(self.mapped_registers):
            dtype = channel["dtype"]
            tmp_monitor_data = drive.monitoring_channel_data(ch_idx, REG_DTYPE(dtype))
            data_array[ch_idx] += tmp_monitor_data

    def stop_reading_data(self) -> None:
        """Stops read_monitoring_data function."""
        self._read_process_finished = True

    @abstractmethod
    def rearm_monitoring(self) -> None:
        """Rearm monitoring."""
        pass

    @abstractmethod
    def _check_buffer_size_is_enough(
        self,
        total_samples: int,
        trigger_delay_samples: int,
        registers: List[Dict[str, Union[int, str, REG_DTYPE]]],
    ) -> None:
        pass

    def _check_samples_and_max_size(
        self, n_sample: int, max_size: int, registers: List[Dict[str, Union[int, str, REG_DTYPE]]]
    ) -> None:
        size_demand = sum(
            self._data_type_size[register["dtype"]] * n_sample
            for register in registers
            if isinstance(register["dtype"], REG_DTYPE)
        )
        if max_size < size_demand:
            raise IMMonitoringError(
                "Number of samples is too high or mapped registers are too big. "
                "Demanded size: {} bytes, buffer max size: {} bytes.".format(size_demand, max_size)
            )
        self.logger.debug(
            "Demanded size: %d bytes, buffer max size: %d bytes.", size_demand, max_size
        )

    def get_trigger_type(self) -> Union[MonitoringSoCType, int]:
        """Get monitoring trigger type.

        Returns:
            Trigger type

        Raises:
            TypeError: If some read value has a wrong type.

        """
        register_value = self.mc.communication.get_register(
            self.MONITOR_START_CONDITION_TYPE_REGISTER, servo=self.servo, axis=0
        )
        if not isinstance(register_value, int):
            raise TypeError("Monitoring trigger type register value has to be an integer")
        try:
            return MonitoringSoCType(register_value)
        except ValueError:
            return register_value

    def raise_forced_trigger(self, blocking: bool = False, timeout: float = 5) -> bool:
        """Raise trigger for Forced Trigger type.

        Args:
            blocking : if ``True``, functions wait until trigger is forced
                (or until the timeout) If ``False``, function try to raise the
                trigger only once.
            timeout : blocking timeout in seconds. ``5`` by default.

        Returns:
            Return ``True`` if trigger is raised, else ``False``.

        """
        trigger_mode = self.get_trigger_type()
        if trigger_mode != MonitoringSoCType.TRIGGER_EVENT_FORCED:
            raise IMMonitoringError("Monitoring trigger type is not Forced Trigger")
        mon_process_stage = None
        final_time = time.time() + timeout
        while mon_process_stage is None or (
            blocking
            and final_time > time.time()
            and mon_process_stage != MonitoringProcessStage.WAITING_FOR_TRIGGER
        ):
            mon_process_stage = self.mc.capture.get_monitoring_process_stage(
                servo=self.servo, version=self._version
            )
            self.mc.communication.set_register(
                self.MONITORING_FORCE_TRIGGER_REGISTER, 1, servo=self.servo, axis=0
            )
        return mon_process_stage >= MonitoringProcessStage.WAITING_FOR_TRIGGER

    def read_monitoring_data_forced_trigger(
        self, trigger_timeout: float = 5
    ) -> List[List[Union[int, float]]]:
        """Trigger and read Forced Trigger monitoring.

        Args:
            trigger_timeout : maximum time function wait to raise the trigger,
                in seconds. ``5`` by default.

        Returns:
            Data of monitoring. Each element of the list is a different register data.

        """
        is_triggered = self.raise_forced_trigger(blocking=True, timeout=trigger_timeout)
        if not is_triggered:
            self.logger.warning("Timeout. Forced trigger is not raised.")
            return [[] for _ in self.mapped_registers]
        return self.read_monitoring_data()



import json
import random
import time

import pytest
from packaging import version

from ingenialink.ethercat.network import EthercatNetwork
from ingenialink.exceptions import ILWrongWorkingCount
from ingenialink.pdo import RPDOMap, RPDOMapItem, TPDOMap, TPDOMapItem
from ingeniamotion.enums import COMMUNICATION_TYPE, OperationMode
from ingeniamotion.exceptions import IMException


@pytest.fixture
def connect_to_all_slaves(motion_controller, pytestconfig):
    mc, alias = motion_controller
    aliases = [alias]
    protocol = pytestconfig.getoption("--protocol")
    if protocol != "soem":
        raise AssertionError("Fixture only available for the soem protocol.")
    config = "tests/config.json"
    with open(config, "r") as fp:
        contents = json.load(fp)
    protocol_contents = contents[protocol]
    connected_slave_id = mc.servos[alias].slave_id
    for slave_content in protocol_contents:
        if slave_content["slave"] == connected_slave_id:
            continue
        alias = f"test{slave_content['slave']}"
        aliases.append(alias)
        mc.communication.connect_servo_ethercat_interface_index(
            slave_content["index"],
            slave_content["slave"],
            slave_content["dictionary"],
            alias,
        )
    yield mc, aliases
    aliases.pop(0)
    for alias in aliases:
        mc.communication.disconnect(alias)


@pytest.mark.soem
def test_create_rpdo_item(motion_controller):
    mc, alias = motion_controller
    position_set_point_initial_value = 100
    position_set_point = mc.capture.pdo.create_pdo_item(
        "CL_POS_SET_POINT_VALUE", servo=alias, value=position_set_point_initial_value
    )
    assert isinstance(position_set_point, RPDOMapItem)
    assert position_set_point.value == position_set_point_initial_value


@pytest.mark.soem
def test_create_tpdo_item(motion_controller):
    mc, alias = motion_controller
    actual_position = mc.capture.pdo.create_pdo_item("CL_POS_FBK_VALUE", servo=alias)
    assert isinstance(actual_position, TPDOMapItem)


@pytest.mark.soem
def test_create_rpdo_item_no_initial_value(motion_controller):
    mc, alias = motion_controller
    with pytest.raises(AttributeError):
        mc.capture.pdo.create_pdo_item("CL_POS_SET_POINT_VALUE", servo=alias)


@pytest.mark.soem
def test_create_empty_rpdo_map(motion_controller):
    mc, alias = motion_controller
    rpdo_map = mc.capture.pdo.create_empty_rpdo_map()
    assert isinstance(rpdo_map, RPDOMap)
    assert len(rpdo_map.items) == 0


@pytest.mark.soem
def test_create_empty_tpdo_map(motion_controller):
    mc, alias = motion_controller
    tpdo_map = mc.capture.pdo.create_empty_tpdo_map()
    assert isinstance(tpdo_map, TPDOMap)
    assert len(tpdo_map.items) == 0


@pytest.mark.soem
def test_create_pdo_maps_single_item(motion_controller):
    mc, alias = motion_controller
    position_set_point = mc.capture.pdo.create_pdo_item(
        "CL_POS_SET_POINT_VALUE", servo=alias, value=0
    )
    actual_position = mc.capture.pdo.create_pdo_item("CL_POS_FBK_VALUE", servo=alias)
    rpdo_map, tpdo_map = mc.capture.pdo.create_pdo_maps(position_set_point, actual_position)
    assert isinstance(rpdo_map, RPDOMap)
    assert len(rpdo_map.items) == 1
    assert position_set_point in rpdo_map.items
    assert isinstance(tpdo_map, TPDOMap)
    assert len(tpdo_map.items) == 1
    assert actual_position in tpdo_map.items


@pytest.mark.soem
def test_create_pdo_maps_list_items(motion_controller):
    mc, alias = motion_controller
    rpdo_regs = ["CL_POS_SET_POINT_VALUE", "CL_VEL_SET_POINT_VALUE"]
    tpdo_regs = ["CL_POS_FBK_VALUE", "CL_VEL_FBK_VALUE"]
    rpdo_items = [
        mc.capture.pdo.create_pdo_item(rpdo_reg, servo=alias, value=0) for rpdo_reg in rpdo_regs
    ]
    tpdo_items = [mc.capture.pdo.create_pdo_item(tpdo_reg, servo=alias) for tpdo_reg in tpdo_regs]
    rpdo_map, tpdo_map = mc.capture.pdo.create_pdo_maps(rpdo_items, tpdo_items)
    assert isinstance(rpdo_map, RPDOMap)
    assert len(rpdo_map.items) == len(rpdo_items)
    assert isinstance(tpdo_map, TPDOMap)
    assert len(tpdo_map.items) == len(tpdo_items)


@pytest.mark.soem
@pytest.mark.parametrize(
    "register, value, pdo_type",
    [
        ("CL_POS_FBK_VALUE", None, "tpdo"),
        ("CL_VEL_SET_POINT_VALUE", 0, "rpdo"),
    ],
)
def test_add_pdo_item_to_map(motion_controller, register, value, pdo_type):
    mc, alias = motion_controller
    if pdo_type == "rpdo":
        pdo_map = mc.capture.pdo.create_empty_rpdo_map()
    else:
        pdo_map = mc.capture.pdo.create_empty_tpdo_map()
    pdo_map_item = mc.capture.pdo.create_pdo_item(register, servo=alias, value=value)
    mc.capture.pdo.add_pdo_item_to_map(pdo_map_item, pdo_map)
    assert len(pdo_map.items) == 1
    assert pdo_map_item in pdo_map.items


@pytest.mark.soem
@pytest.mark.parametrize(
    "register, value, pdo_type",
    [
        ("CL_POS_FBK_VALUE", None, "tpdo"),
        ("CL_VEL_SET_POINT_VALUE", 0, "rpdo"),
    ],
)
def test_add_pdo_item_to_map_exceptions(motion_controller, register, value, pdo_type):
    mc, alias = motion_controller
    if pdo_type == "tpdo":
        pdo_map = mc.capture.pdo.create_empty_rpdo_map()
    else:
        pdo_map = mc.capture.pdo.create_empty_tpdo_map()
    pdo_map_item = mc.capture.pdo.create_pdo_item(register, servo=alias, value=value)
    with pytest.raises(ValueError):
        mc.capture.pdo.add_pdo_item_to_map(pdo_map_item, pdo_map)


@pytest.mark.soem
@pytest.mark.parametrize(
    "rpdo_maps, tpdo_maps",
    [
        [["rpdo", "rpdo"], ["rpdo", "tpdo"]],
        [["rpdo", "tpdo"], ["tpdo", "tpdo"]],
    ],
)
def test_set_pdo_maps_to_slave_exception(motion_controller, rpdo_maps, tpdo_maps):
    mc, alias = motion_controller
    rx_maps = []
    tx_maps = []
    for map_type in rpdo_maps:
        pdo_map = (
            mc.capture.pdo.create_empty_rpdo_map()
            if map_type == "rpdo"
            else mc.capture.pdo.create_empty_tpdo_map()
        )
        rx_maps.append(pdo_map)
    for map_type in tpdo_maps:
        pdo_map = (
            mc.capture.pdo.create_empty_rpdo_map()
            if map_type == "rpdo"
            else mc.capture.pdo.create_empty_tpdo_map()
        )
        tx_maps.append(pdo_map)
    with pytest.raises(ValueError):
        mc.capture.pdo.set_pdo_maps_to_slave(rx_maps, tx_maps, alias)


@pytest.mark.parametrize(
    "refresh_rate",
    [
        0.0001,
        5,
    ],
)
@pytest.mark.soem
def test_pdos_refresh_rate(motion_controller, refresh_rate):
    mc, alias = motion_controller
    with pytest.raises(ValueError):
        mc.capture.pdo.start_pdos(COMMUNICATION_TYPE.Ethercat, refresh_rate)


@pytest.mark.soem
def test_start_pdos(connect_to_all_slaves):
    mc, aliases = connect_to_all_slaves
    pdo_map_items = {}
    initial_operation_modes = {}
    rpdo_values = {}
    tpdo_values = {}
    for alias in aliases:
        rpdo_map = mc.capture.pdo.create_empty_rpdo_map()
        tpdo_map = mc.capture.pdo.create_empty_tpdo_map()
        initial_operation_mode = mc.motion.get_operation_mode(servo=alias)
        operation_mode = mc.capture.pdo.create_pdo_item(
            "DRV_OP_CMD", servo=alias, value=initial_operation_mode.value
        )
        actual_position = mc.capture.pdo.create_pdo_item("CL_POS_FBK_VALUE", servo=alias)
        mc.capture.pdo.add_pdo_item_to_map(operation_mode, rpdo_map)
        mc.capture.pdo.add_pdo_item_to_map(actual_position, tpdo_map)
        mc.capture.pdo.set_pdo_maps_to_slave(rpdo_map, tpdo_map, servo=alias)
        pdo_map_items[alias] = (operation_mode, actual_position)
        random_op_mode = random.choice(
            [op_mode for op_mode in OperationMode if op_mode != initial_operation_mode]
        )
        initial_operation_modes[alias] = initial_operation_mode
        rpdo_values[alias] = random_op_mode

    def send_callback():
        for alias in aliases:
            rpdo_map_item, _ = pdo_map_items[alias]
            rpdo_map_item.value = rpdo_values[alias].value

    def receive_callback():
        for alias in aliases:
            _, tpdo_map_item = pdo_map_items[alias]
            tpdo_values[alias] = tpdo_map_item.value

    mc.capture.pdo.subscribe_to_send_process_data(send_callback)
    mc.capture.pdo.subscribe_to_receive_process_data(receive_callback)
    refresh_rate = 0.5
    mc.capture.pdo.start_pdos(refresh_rate=refresh_rate)
    time.sleep(2 * refresh_rate)
    mc.capture.pdo.stop_pdos()
    for alias in aliases:
        # Check that RPDO are being sent
        assert rpdo_values[alias] == mc.motion.get_operation_mode(servo=alias)
        # Check that TPDO are being received
        assert pytest.approx(tpdo_values[alias], abs=2) == mc.motion.get_actual_position(
            servo=alias
        )
        # Restore the initial operation mode
        mc.motion.set_operation_mode(initial_operation_modes[alias], servo=alias)
        mc.capture.pdo.remove_rpdo_map(alias, rpdo_map_index=0)
        mc.capture.pdo.remove_tpdo_map(alias, tpdo_map_index=0)


@pytest.mark.soem
def test_stop_pdos_exception(motion_controller):
    mc, alias = motion_controller
    with pytest.raises(IMException):
        mc.capture.pdo.stop_pdos()


@pytest.mark.soem
def test_start_pdos_not_implemented_exception(motion_controller):
    mc, alias = motion_controller
    with pytest.raises(NotImplementedError):
        mc.capture.pdo.start_pdos(COMMUNICATION_TYPE.Canopen)


@pytest.mark.soem
def test_start_pdos_wrong_network_type_exception(motion_controller):
    mc, alias = motion_controller
    with pytest.raises(ValueError):
        mc.capture.pdo.start_pdos("ethernet")


@pytest.mark.soem
def test_start_pdos_number_of_network_exception(mocker, motion_controller):
    mc, alias = motion_controller
    mock_net = {"ifname1": EthercatNetwork("ifname1"), "ifname2": EthercatNetwork("ifname2")}
    mocker.patch.object(mc, "_MotionController__net", mock_net)
    with pytest.raises(ValueError):
        mc.capture.pdo.start_pdos()
    with pytest.raises(IMException):
        mc.capture.pdo.start_pdos(COMMUNICATION_TYPE.Ethercat)


def skip_if_pdo_padding_is_not_available(mc, alias):
    pdo_poller_fw_version = "2.5.0"
    firmware_version = mc.configuration.get_fw_version(alias, 1)
    if version.parse(firmware_version) < version.parse(pdo_poller_fw_version):
        pytest.skip(
            f"PDO poller is available for firmware version {pdo_poller_fw_version} or higher. "
            f"Firmware version found: {firmware_version}"
        )


@pytest.mark.soem
def test_create_poller(motion_controller):
    mc, alias = motion_controller
    skip_if_pdo_padding_is_not_available(mc, alias)
    registers = [{"name": "CL_POS_FBK_VALUE", "axis": 1}, {"name": "CL_VEL_FBK_VALUE", "axis": 1}]
    sampling_time = 0.25
    sleep_time = 1
    poller = mc.capture.pdo.create_poller(registers, alias, sampling_time)
    time.sleep(sleep_time)
    poller.stop()
    timestamps, data = poller.data
    channel_0_data, channel_1_data = data
    assert len(channel_0_data) == sleep_time / sampling_time
    assert len(channel_0_data) == len(channel_1_data)
    assert len(data) == len(registers)
    assert len(timestamps) == len(channel_0_data)


def dummy_callback(e):
    pass


@pytest.mark.soem
def test_subscribe_exceptions(motion_controller, mocker):
    mc, _ = motion_controller

    error_msg = "Test error"

    def send_receive_processdata(self, *args):
        raise ILWrongWorkingCount(error_msg)

    mocker.patch("ingenialink.ethercat.network.EthercatNetwork.start_pdos")
    mocker.patch("ingenialink.ethercat.network.EthercatNetwork.stop_pdos")
    mocker.patch(
        "ingenialink.ethercat.network.EthercatNetwork.send_receive_processdata",
        new=send_receive_processdata,
    )
    patch_callback = mocker.patch("tests.test_communication.dummy_callback")

    mc.capture.pdo.subscribe_to_exceptions(patch_callback)
    mc.capture.pdo.start_pdos()

    t = time.time()
    timeout = 5
    while not mc.capture.pdo._pdo_thread._pd_thread_stop_event.is_set() and (
        (time.time() - t) < timeout
    ):
        pass

    assert mc.capture.pdo._pdo_thread._pd_thread_stop_event.is_set()
    patch_callback.assert_called_once()
    assert (
        str(patch_callback.call_args_list[0][0][0])
        == f"Stopping the PDO thread due to the following exception: {error_msg} "
    )



from typing import TYPE_CHECKING, Dict, Union

import ingenialogger
from ingenialink.exceptions import ILError

from ingeniamotion.wizard_tests.base_test import BaseTest
from ingeniamotion.wizard_tests.stoppable import StopException
from ingeniamotion.enums import OperationMode, SeverityLevel, CommutationMode
from ingeniamotion.metaclass import DEFAULT_SERVO, DEFAULT_AXIS

if TYPE_CHECKING:
    from ingeniamotion import MotionController


class Brake(BaseTest):

    BRAKE_OVERRIDE_REGISTER = "MOT_BRAKE_OVERRIDE"

    BACKUP_REGISTERS = [
        "MOT_BRAKE_OVERRIDE",
        "DRV_OP_CMD",
        "MOT_PAIR_POLES",
        "COMMU_PHASING_MODE",
        "COMMU_ANGLE_SENSOR",
        "MOT_COMMU_MOD",
    ]

    def __init__(
        self, mc: "MotionController", servo: str = DEFAULT_SERVO, axis: int = DEFAULT_AXIS
    ) -> None:
        super().__init__()
        self.mc = mc
        self.servo = servo
        self.axis = axis
        self.logger = ingenialogger.get_logger(__name__, axis=axis, drive=mc.servo_name(servo))
        self.backup_registers_names = self.BACKUP_REGISTERS

    def setup(self) -> None:
        self.mc.motion.motor_disable(servo=self.servo, axis=self.axis)
        self.mc.configuration.disable_brake_override(servo=self.servo, axis=self.axis)
        self.mc.configuration.set_commutation_mode(
            CommutationMode.SINUSOIDAL, servo=self.servo, axis=self.axis
        )
        self.mc.motion.set_internal_generator_configuration(
            OperationMode.VOLTAGE, servo=self.servo, axis=self.axis
        )

    def loop(self) -> None:
        self.mc.motion.motor_enable(servo=self.servo, axis=self.axis)

    def teardown(self) -> None:
        self.mc.motion.motor_disable(servo=self.servo, axis=self.axis)

    def finish(self) -> Dict[str, Union[SeverityLevel, str]]:
        try:
            self.teardown()
        finally:
            self.restore_backup_registers()
        output = SeverityLevel.SUCCESS
        return {
            "result_severity": self.get_result_severity(output),
            "result_message": self.get_result_msg(output),
        }

    def run(self) -> None:
        self.reset_stop()
        self.save_backup_registers()
        try:
            self.setup()
            self.loop()
        except ILError as err:
            self.finish()
            raise err
        except StopException:
            self.logger.warning("Test has been stopped")
            self.finish()

    def get_result_severity(self, output: SeverityLevel) -> SeverityLevel:
        return output

    def get_result_msg(self, output: SeverityLevel) -> str:
        if output == SeverityLevel.SUCCESS:
            return "Success"
        else:
            return "Fail"



import time

import pytest

from threading import Thread
from functools import partial

from ingeniamotion.enums import MonitoringSoCType, MonitoringSoCConfig
from ingeniamotion.exceptions import IMMonitoringError

MONITOR_START_CONDITION_TYPE_REGISTER = "MON_CFG_SOC_TYPE"


class ThreadWithReturnValue(Thread):
    def __init__(self, *init_args, **init_kwargs):
        Thread.__init__(self, *init_args, **init_kwargs)
        self._return = None

    def run(self):
        self._return = self._target(*self._args, **self._kwargs)

    def join(self):
        Thread.join(self)
        return self._return


@pytest.fixture
def monitoring(skip_if_monitoring_not_available, motion_controller):
    mc, alias = motion_controller
    return mc.capture.create_empty_monitoring(alias)


@pytest.fixture
def mon_set_freq(skip_if_monitoring_not_available, monitoring):
    monitoring.set_frequency(10)


@pytest.fixture
def mon_map_registers(skip_if_monitoring_not_available, monitoring):
    monitoring.map_registers([{"axis": 1, "name": "CL_POS_FBK_VALUE"}])


@pytest.mark.virtual
@pytest.mark.smoke
@pytest.mark.parametrize(
    "trigger_type",
    [
        MonitoringSoCType.TRIGGER_EVENT_AUTO,
        MonitoringSoCType.TRIGGER_EVENT_FORCED,
        MonitoringSoCType.TRIGGER_EVENT_EDGE,
        4,  # Invalid value
    ],
)
def test_get_trigger_type(motion_controller, monitoring, trigger_type):
    mc, alias = motion_controller
    mc.communication.set_register(
        MONITOR_START_CONDITION_TYPE_REGISTER, trigger_type, servo=alias, axis=0
    )
    test_trigger = monitoring.get_trigger_type()
    assert test_trigger == trigger_type


@pytest.mark.parametrize(
    "block, timeout, sample_t, wait, result",
    [
        (False, 5, 0.8, 2, True),
        (True, 6, 0.8, 0, True),
        (False, 5, 0.8, 0, False),
        (True, 0.1, 0.8, 0, False),
    ],
)
def test_raise_forced_trigger(
    motion_controller,
    monitoring,
    block,
    timeout,
    sample_t,
    wait,
    result,
    mon_set_freq,
    mon_map_registers,
    disable_monitoring_disturbance,
):
    mc, alias = motion_controller
    monitoring.set_trigger(MonitoringSoCType.TRIGGER_EVENT_FORCED)
    monitoring.configure_sample_time(sample_t, 0)
    mc.capture.enable_monitoring_disturbance(servo=alias)
    time.sleep(wait)
    test_output = monitoring.raise_forced_trigger(blocking=block, timeout=timeout)
    assert test_output == result
    time.sleep(1)


@pytest.mark.smoke
@pytest.mark.usefixtures("mon_set_freq")
@pytest.mark.usefixtures("mon_map_registers")
def test_raise_forced_trigger_fail(motion_controller, monitoring, disable_monitoring_disturbance):
    mc, alias = motion_controller
    monitoring.set_trigger(MonitoringSoCType.TRIGGER_EVENT_AUTO)
    monitoring.configure_sample_time(0.8, 0)
    mc.capture.enable_monitoring_disturbance(servo=alias)
    with pytest.raises(IMMonitoringError):
        monitoring.raise_forced_trigger()


@pytest.mark.usefixtures("mon_set_freq")
@pytest.mark.usefixtures("mon_map_registers")
@pytest.mark.parametrize(
    "timeout, sample_t, result",
    [
        (5, 0.8, True),
        (0.1, 0.8, False),
    ],
)
def test_read_monitoring_data_forced_trigger(
    motion_controller, monitoring, timeout, sample_t, result, disable_monitoring_disturbance
):
    mc, alias = motion_controller
    monitoring.set_trigger(MonitoringSoCType.TRIGGER_EVENT_FORCED)
    monitoring.configure_sample_time(sample_t, 0)
    mc.capture.enable_monitoring_disturbance(servo=alias)
    test_output = monitoring.read_monitoring_data_forced_trigger(timeout)
    if result:
        assert len(test_output[0]) == monitoring.samples_number
    else:
        assert len(test_output[0]) == 0


@pytest.mark.smoke
@pytest.mark.parametrize("prescaler", list(range(2, 11, 2)))
def test_set_monitoring_frequency(motion_controller, monitoring, prescaler):
    mc, alias = motion_controller
    monitoring.set_frequency(prescaler)
    value = mc.communication.get_register(
        monitoring.MONITORING_FREQUENCY_DIVIDER_REGISTER, servo=alias, axis=0
    )
    assert value == prescaler


@pytest.mark.smoke
def test_set_monitoring_frequency_exception(monitoring):
    prescaler = 0.5
    with pytest.raises(ValueError):
        monitoring.set_frequency(prescaler)


@pytest.mark.virtual
@pytest.mark.smoke
def test_monitoring_map_registers_size_exception(monitoring):
    registers = [{"axis": 1, "name": "CL_POS_FBK_VALUE"}]
    monitoring.samples_number = monitoring.max_sample_number
    with pytest.raises(IMMonitoringError):
        monitoring.map_registers(registers)


@pytest.mark.virtual
@pytest.mark.smoke
def test_monitoring_map_registers_fail(monitoring):
    registers = []
    with pytest.raises(IMMonitoringError):
        monitoring.map_registers(registers)


@pytest.mark.smoke
@pytest.mark.usefixtures("mon_map_registers")
@pytest.mark.parametrize(
    "trigger_type, edge_condition, trigger_signal, trigger_value",
    [
        (MonitoringSoCType.TRIGGER_EVENT_AUTO, None, None, None),
        (MonitoringSoCType.TRIGGER_EVENT_FORCED, None, None, None),
        (
            MonitoringSoCType.TRIGGER_EVENT_EDGE,
            MonitoringSoCConfig.TRIGGER_CONFIG_RISING,
            {"axis": 1, "name": "CL_POS_FBK_VALUE"},
            0.5,
        ),
    ],
)
def test_monitoring_set_trigger(
    motion_controller, monitoring, trigger_type, edge_condition, trigger_signal, trigger_value
):
    mc, alias = motion_controller
    monitoring.set_trigger(trigger_type, edge_condition, trigger_signal, trigger_value)
    value = mc.communication.get_register(
        MONITOR_START_CONDITION_TYPE_REGISTER, servo=alias, axis=0
    )
    assert value == trigger_type


@pytest.mark.smoke
@pytest.mark.usefixtures("mon_map_registers")
@pytest.mark.parametrize(
    "trigger_type, edge_condition, trigger_signal, trigger_value",
    [
        (MonitoringSoCType.TRIGGER_EVENT_EDGE, None, {"axis": 1, "name": "CL_POS_FBK_VALUE"}, 0.5),
        (
            MonitoringSoCType.TRIGGER_EVENT_EDGE,
            MonitoringSoCConfig.TRIGGER_CONFIG_RISING,
            None,
            0.5,
        ),
        (
            MonitoringSoCType.TRIGGER_EVENT_EDGE,
            MonitoringSoCConfig.TRIGGER_CONFIG_RISING,
            {"axis": 1, "name": "CL_POS_FBK_VALUE"},
            None,
        ),
    ],
)
def test_monitoring_set_trigger_exceptions(
    monitoring, trigger_type, edge_condition, trigger_signal, trigger_value
):
    with pytest.raises(TypeError):
        monitoring.set_trigger(trigger_type, edge_condition, trigger_signal, trigger_value)


@pytest.mark.smoke
def test_configure_number_samples(motion_controller, monitoring):
    mc, alias = motion_controller
    total_num_samples = 500
    trigger_delay_samples = 100
    monitoring.configure_number_samples(total_num_samples, trigger_delay_samples)
    value = mc.communication.get_register(
        monitoring.MONITORING_WINDOW_NUMBER_SAMPLES_REGISTER, servo=alias, axis=0
    )
    assert value == total_num_samples
    value = mc.communication.get_register(
        monitoring.MONITORING_TRIGGER_DELAY_SAMPLES_REGISTER, servo=alias, axis=0
    )
    assert value == trigger_delay_samples


@pytest.mark.smoke
@pytest.mark.parametrize("total_num_samples, trigger_delay_samples", [(500, 510), (510, -500)])
def test_configure_number_samples_exceptions(monitoring, total_num_samples, trigger_delay_samples):
    with pytest.raises(ValueError):
        monitoring.configure_number_samples(total_num_samples, trigger_delay_samples)


@pytest.mark.smoke
def test_configure_sample_time(motion_controller, monitoring):
    mc, alias = motion_controller
    total_time = 5
    sampling_freq = 1e4
    monitoring.sampling_freq = sampling_freq
    trigger_delay = total_time // 2
    total_num_samples = int(sampling_freq * total_time)
    trigger_delay_samples = int(((total_time / 2) - trigger_delay) * sampling_freq)
    monitoring.configure_sample_time(total_time, trigger_delay)
    value = mc.communication.get_register(
        monitoring.MONITORING_WINDOW_NUMBER_SAMPLES_REGISTER, servo=alias, axis=0
    )
    assert value == total_num_samples
    value = mc.communication.get_register(
        monitoring.MONITORING_TRIGGER_DELAY_SAMPLES_REGISTER, servo=alias, axis=0
    )
    assert value == trigger_delay_samples


@pytest.mark.smoke
@pytest.mark.parametrize("total_time, sign", [(5, 1), (5, -1)])
def test_configure_sample_time_exception(monitoring, total_time, sign):
    trigger_delay = sign * ((total_time // 2) + 1)
    with pytest.raises(ValueError):
        monitoring.configure_sample_time(total_time, trigger_delay)


@pytest.mark.skip("Check if channels are configured is not implemented yet")
@pytest.mark.usefixtures("disable_monitoring_disturbance")
def test_read_monitoring_data_not_configured(motion_controller, monitoring):
    # TODO Add exception in function for this test case
    mc, alias = motion_controller
    drive = mc._get_drive(alias)
    drive.monitoring_remove_all_mapped_registers()
    mc.capture.enable_monitoring_disturbance(servo=alias)
    monitoring.samples_number = monitoring.max_sample_number
    test_output = monitoring.read_monitoring_data()
    assert len(test_output[0]) == 0


@pytest.mark.smoke
@pytest.mark.usefixtures("mon_set_freq")
@pytest.mark.usefixtures("mon_map_registers")
def test_read_monitoring_data_disabled(monitoring):
    # TODO Add exception in function for this test case
    monitoring.configure_sample_time(0.8, 0)
    test_output = monitoring.read_monitoring_data()
    assert len(test_output[0]) == 0


@pytest.mark.usefixtures("mon_set_freq")
@pytest.mark.usefixtures("mon_map_registers")
def test_read_monitoring_data_timeout(
    motion_controller, monitoring, disable_monitoring_disturbance
):
    timeout = 2
    sample_t = 0.8
    mc, alias = motion_controller
    monitoring.set_trigger(MonitoringSoCType.TRIGGER_EVENT_FORCED)
    monitoring.configure_sample_time(sample_t, 0)
    mc.capture.enable_monitoring_disturbance(servo=alias)
    test_output = monitoring.read_monitoring_data(timeout)
    assert len(test_output[0]) == 0


@pytest.mark.usefixtures("mon_set_freq")
@pytest.mark.usefixtures("mon_map_registers")
def test_read_monitoring_data_no_rearm(
    motion_controller, monitoring, disable_monitoring_disturbance
):
    sample_t = 0.8
    timeout = 2
    block = True
    wait = 2
    mc, alias = motion_controller
    monitoring.set_trigger(MonitoringSoCType.TRIGGER_EVENT_FORCED)
    monitoring.configure_sample_time(sample_t, 0)
    mc.capture.enable_monitoring_disturbance(servo=alias)
    time.sleep(wait)
    trigger_raised = monitoring.raise_forced_trigger(block, timeout)
    assert trigger_raised
    test_output = monitoring.read_monitoring_data()
    assert len(test_output[0]) > 0
    trigger_raised = monitoring.raise_forced_trigger(block, timeout)
    assert not trigger_raised
    test_output = monitoring.read_monitoring_data()
    assert len(test_output[0]) == 0


@pytest.mark.usefixtures("mon_set_freq")
@pytest.mark.usefixtures("mon_map_registers")
def test_rearm_monitoring(motion_controller, monitoring, disable_monitoring_disturbance):
    sample_t = 0.8
    timeout = 2
    block = True
    wait = 2
    mc, alias = motion_controller
    monitoring.set_trigger(MonitoringSoCType.TRIGGER_EVENT_FORCED)
    monitoring.configure_sample_time(sample_t, 0)
    mc.capture.enable_monitoring_disturbance(servo=alias)
    time.sleep(wait)
    for _ in range(3):
        trigger_raised = monitoring.raise_forced_trigger(block, timeout)
        assert trigger_raised
        test_output = monitoring.read_monitoring_data()
        assert len(test_output[0]) > 0
        monitoring.rearm_monitoring()
        time.sleep(wait // 2)


def run_read_monitoring_data_and_stop(monitoring, timeout):
    read_monitoring_data_timeout = partial(monitoring.read_monitoring_data, timeout=timeout)
    test_thread = ThreadWithReturnValue(target=read_monitoring_data_timeout)
    test_thread.start()
    monitoring.stop_reading_data()
    return test_thread.join()


@pytest.mark.smoke
@pytest.mark.usefixtures("mon_set_freq")
@pytest.mark.usefixtures("mon_map_registers")
def test_stop_reading_data(motion_controller, monitoring, disable_monitoring_disturbance):
    sample_t = 0.8
    timeout = 10
    mc, alias = motion_controller
    monitoring.set_trigger(MonitoringSoCType.TRIGGER_EVENT_FORCED)
    monitoring.configure_sample_time(sample_t, 0)
    mc.capture.enable_monitoring_disturbance(servo=alias)
    init_time = time.time()
    test_output = run_read_monitoring_data_and_stop(monitoring, timeout)
    assert len(test_output[0]) == 0
    assert (time.time() - init_time) < timeout


@pytest.mark.virtual
@pytest.mark.smoke
def test_monitoring_max_sample_size(motion_controller, skip_if_monitoring_not_available):
    mc, alias = motion_controller
    target_register = mc.capture.MONITORING_MAXIMUM_SAMPLE_SIZE_REGISTER
    axis = 0
    max_sample_size = mc.capture.monitoring_max_sample_size(servo=alias)
    drive = mc.servos[alias]
    value = drive.read(target_register, subnode=axis)
    assert max_sample_size == value


@pytest.mark.virtual
@pytest.mark.smoke
def test_monitoring_map_registers_invalid_subnode(mocker, motion_controller, monitoring):
    registers = [{"axis": "1", "name": "DRV_AXIS_NUMBER"}]
    mc, alias = motion_controller
    mocker.patch.object(mc.capture, "is_monitoring_enabled", return_value=False)
    with pytest.raises(TypeError):
        monitoring.map_registers(registers)


@pytest.mark.virtual
@pytest.mark.smoke
def test_monitoring_map_registers_invalid_register(mocker, motion_controller, monitoring):
    registers = [{"axis": 1, "name": 1}]
    mc, alias = motion_controller
    mocker.patch.object(mc.capture, "is_monitoring_enabled", return_value=False)
    with pytest.raises(TypeError):
        monitoring.map_registers(registers)


@pytest.mark.virtual
@pytest.mark.smoke
def test_monitoring_map_registers_invalid_number_mapped_registers(
    mocker, motion_controller, monitoring
):
    registers = [{"axis": 1, "name": "CL_POS_FBK_VALUE"}]
    mc, alias = motion_controller
    mocker.patch.object(mc.capture, "is_monitoring_enabled", return_value=False)
    mocker.patch.object(mc.communication, "get_register", return_value="invalid_value")
    with pytest.raises(TypeError):
        monitoring.map_registers(registers)


@pytest.mark.parametrize(
    "trigger_delay, expected_exception",
    [
        (3, ValueError),
        (2.5, TypeError),
    ],
)
@pytest.mark.virtual
@pytest.mark.smoke
def test_configure_sample_time_exceptions(
    mocker, motion_controller, monitoring, trigger_delay, expected_exception
):
    mc, alias = motion_controller
    total_time = 5
    mocker.patch.object(mc.capture, "is_monitoring_enabled", return_value=False)
    with pytest.raises(expected_exception):
        monitoring.configure_sample_time(total_time, trigger_delay)


@pytest.mark.virtual
def test_get_trigger_type_exception(mocker, motion_controller, monitoring):
    mc, alias = motion_controller
    mocker.patch.object(mc.communication, "get_register", return_value="invalid_value")
    with pytest.raises(TypeError):
        monitoring.get_trigger_type()



from enum import IntEnum
from typing import Dict

from ingenialink.network import Network
from ingenialink.servo import Servo

from ingeniamotion.capture import Capture
from ingeniamotion.communication import Communication
from ingeniamotion.configuration import Configuration
from ingeniamotion.drive_tests import DriveTests
from ingeniamotion.errors import Errors
from ingeniamotion.information import Information
from ingeniamotion.metaclass import DEFAULT_AXIS, DEFAULT_SERVO
from ingeniamotion.motion import Motion


class MotionController:
    """Motion Controller."""

    def __init__(self) -> None:
        self.__servos: Dict[str, Servo] = {}
        self.__net: Dict[str, Network] = {}
        self.__servo_net: Dict[str, str] = {}
        self.__config: Configuration = Configuration(self)
        self.__motion: Motion = Motion(self)
        self.__capture: Capture = Capture(self)
        self.__comm: Communication = Communication(self)
        self.__tests: DriveTests = DriveTests(self)
        self.__errors: Errors = Errors(self)
        self.__info: Information = Information(self)

    def servo_name(self, servo: str = DEFAULT_SERVO) -> str:
        return "{} ({})".format(self.servos[servo].info["product_code"], servo)

    def get_register_enum(
        self, register: str, servo: str = DEFAULT_SERVO, axis: int = DEFAULT_AXIS
    ) -> IntEnum:
        drive = self.servos[servo]
        enum_dict = drive.dictionary.registers(axis)[register].enums
        return IntEnum(register, enum_dict)

    def is_alive(self, servo: str = DEFAULT_SERVO) -> bool:
        """Check if the servo is alive.

        Args:
            servo : servo alias to reference it. ``default`` by default.

        Returns:
            ``True`` if the servo is alive, ``False`` otherwise.

        """
        drive = self._get_drive(servo)
        return drive.is_alive()

    def _get_network(self, servo: str) -> Network:
        """Return servo network instance.

        Args:
            servo : servo alias to reference it. ``default`` by default.

        Returns:
            Network instance of the servo.

        """
        net_key = self.servo_net[servo]
        return self.net[net_key]

    def _get_drive(self, servo: str) -> Servo:
        """Return servo drive instance.

        Args:
            servo : servo alias to reference it. ``default`` by default.

        Returns:
            Servo instance.

        """
        return self.servos[servo]

    # Properties
    @property
    def servos(self) -> Dict[str, Servo]:
        """Dict of ``ingenialink.Servo`` connected indexed by alias"""
        return self.__servos

    @servos.setter
    def servos(self, value: Dict[str, Servo]) -> None:
        self.__servos = value

    @property
    def net(self) -> Dict[str, Network]:
        """Dict of ``ingenialink.Network`` connected indexed by alias"""
        return self.__net

    @net.setter
    def net(self, value: Dict[str, Network]) -> None:
        self.__net = value

    @property
    def servo_net(self) -> Dict[str, str]:
        return self.__servo_net

    @servo_net.setter
    def servo_net(self, value: Dict[str, str]) -> None:
        self.__servo_net = value

    @property
    def configuration(self) -> Configuration:
        """Instance of  :class:`~ingeniamotion.configuration.Configuration` class"""
        return self.__config

    @property
    def motion(self) -> Motion:
        """Instance of  :class:`~ingeniamotion.motion.Motion` class"""
        return self.__motion

    @property
    def capture(self) -> Capture:
        """Instance of  :class:`~ingeniamotion.capture.Capture` class"""
        return self.__capture

    @property
    def communication(self) -> Communication:
        """Instance of  :class:`~ingeniamotion.communication.Communication` class"""
        return self.__comm

    @property
    def tests(self) -> DriveTests:
        """Instance of  :class:`~ingeniamotion.drive_tests.DriveTests` class"""
        return self.__tests

    @property
    def errors(self) -> Errors:
        """Instance of :class:`~ingeniamotion.errors.Errors` class"""
        return self.__errors

    @property
    def info(self) -> Information:
        """Instance of :class:`~ingeniamotion.errors.Information` class"""
        return self.__info



from typing import TYPE_CHECKING, List
import ingenialogger

from enum import IntEnum

from ingeniamotion.wizard_tests.base_test import BaseTest
from ingeniamotion.enums import SeverityLevel

if TYPE_CHECKING:
    from ingeniamotion import MotionController


class STOTest(BaseTest):
    """STO test."""

    class ResultType(IntEnum):
        STO_INACTIVE = 0
        STO_ACTIVE = -1
        STO_ABNORMAL_LATCHED = -2
        STO_ABNORMAL = -3
        STO_ABNORMAL_SUPPLY = -4
        STO_INPUTS_DIFFER = -5

    class Polarity(IntEnum):
        NORMAL = 0
        REVERSED = 1

    result_description = {
        ResultType.STO_INACTIVE: "STO Inactive",
        ResultType.STO_ACTIVE: "STO Active",
        ResultType.STO_ABNORMAL_LATCHED: "Abnormal STO Latched",
        ResultType.STO_ABNORMAL: "Abnormal STO",
        ResultType.STO_ABNORMAL_SUPPLY: "Abnormal Supply",
        ResultType.STO_INPUTS_DIFFER: "STO Inputs Differ",
    }

    STO_STATUS_REGISTER = "DRV_PROT_STO_STATUS"

    STO1_ACTIVE_BIT = 0x1
    STO2_ACTIVE_BIT = 0x2
    STO_SUPPLY_FAULT_BIT = 0x4
    STO_ABNORMAL_FAULT_BIT = 0x8
    STO_REPORT_BIT = 0x10

    BACKUP_REGISTERS: List[str] = []

    def __init__(self, mc: "MotionController", servo: str, axis: int) -> None:
        super().__init__()
        self.mc = mc
        self.servo = servo
        self.axis = axis
        self.logger = ingenialogger.get_logger(__name__, axis=axis, drive=mc.servo_name(servo))
        self.backup_registers_names = self.BACKUP_REGISTERS
        self.suggested_registers = {}
        self.TEST_TYPE = self.ResultType

    def setup(self) -> None:
        pass

    def teardown(self) -> None:
        pass

    def loop(self) -> ResultType:
        if self.mc.configuration.is_sto1_active(servo=self.servo, axis=self.axis) == 0:
            self.logger.info("STO1 bit is LOW")
        # Check STO1 status --> Check bit 0 (0x1 in HEX)
        else:
            self.logger.info("STO1 bit is HIGH")

        if self.mc.configuration.is_sto2_active(servo=self.servo, axis=self.axis) == 0:
            self.logger.info("STO2 bit is LOW")
        # Check STO2 status --> Check bit 1 (0x2 in HEX)
        else:
            self.logger.info("STO2 bit is HIGH")

        # Check STO supply fault status --> Check bit 2 (0x4 in HEX)
        sto_power_supply = self.mc.configuration.check_sto_power_supply(
            servo=self.servo, axis=self.axis
        )
        if sto_power_supply == 0:
            self.logger.info("STO Power Supply is LOW")
        else:
            self.logger.info("STO Power Supply is HIGH")

        # Check STO abnormal fault status --> Check bit 3 (0x8 in HEX)
        sto_abnormal_fault = self.mc.configuration.check_sto_abnormal_fault(
            servo=self.servo, axis=self.axis
        )
        if sto_abnormal_fault == 0:
            self.logger.info("STO abnormal fault bit is LOW")
        else:
            self.logger.info("STO abnormal fault bit is HIGH")

        # Check STO report --> Check bit 4 (0x10 in HEX)
        if self.mc.configuration.get_sto_report_bit(servo=self.servo, axis=self.axis) == 0:
            self.logger.info("STO report is LOW")
        else:
            self.logger.info("STO report is HIGH")

        # Check STO STATE
        if self.mc.configuration.is_sto_active(servo=self.servo, axis=self.axis):
            return self.ResultType.STO_ACTIVE
        elif self.mc.configuration.is_sto_inactive(servo=self.servo, axis=self.axis):
            return self.ResultType.STO_INACTIVE
        elif self.mc.configuration.is_sto_abnormal_latched(servo=self.servo, axis=self.axis):
            return self.ResultType.STO_ABNORMAL_LATCHED
        elif sto_abnormal_fault != 0:
            return self.ResultType.STO_ABNORMAL
        elif sto_power_supply == 0:
            return self.ResultType.STO_ABNORMAL_SUPPLY
        else:
            return self.ResultType.STO_INPUTS_DIFFER

    def get_result_msg(self, output: ResultType) -> str:
        return self.result_description[output]

    def get_result_severity(self, output: ResultType) -> SeverityLevel:
        if output < self.ResultType.STO_INACTIVE:
            return SeverityLevel.FAIL
        else:
            return SeverityLevel.SUCCESS



import argparse
import json
import os
import time
from functools import partial

import ifaddr
import ingenialogger
from ingenialink.canopen.network import CAN_BAUDRATE, CAN_DEVICE, CanopenNetwork
from ingenialink.ethercat.network import EthercatNetwork
from ingenialink.ethernet.network import EthernetNetwork
from ingenialink.exceptions import ILError, ILFirmwareLoadError
from ping3 import ping

logger = ingenialogger.get_logger("load_FWs")
ingenialogger.configure_logger(level=ingenialogger.LoggingLevel.INFO)
dirname = os.path.dirname(__file__)


def setup_command():
    parser = argparse.ArgumentParser(description="Run feedback test")
    parser.add_argument("comm", help="communication protocol", choices=["canopen", "soem", "eoe"])
    return parser.parse_args()


def connect_can(drive_conf):
    net = CanopenNetwork(
        CAN_DEVICE(drive_conf["device"]),
        drive_conf["channel"],
        CAN_BAUDRATE(drive_conf["baudrate"]),
    )
    servo = net.connect_to_slave(drive_conf["node_id"], drive_conf["dictionary"])
    return net, servo


def load_can(drive_conf):
    # Number of reattempts for trying the CAN bootloader
    BL_NUM_OF_REATTEMPTS = 4

    # Timings, in seconds
    SLEEP_TIME_AFTER_ATTEMP = 5.0
    SLEEP_TIME_AFTER_BL = 5.0
    TIMEOUT_NEW_FW_DETECT = 30.0
    SLEEP_TIME_NEW_FW_DETECT = 5.0

    for attempt in range(BL_NUM_OF_REATTEMPTS):
        logger.info(f"CAN boot attempt {attempt + 1} of {BL_NUM_OF_REATTEMPTS}")
        try:
            net, servo = connect_can(drive_conf)
            logger.info(
                "Drive connected. %s, node: %d, baudrate: %d, channel: %d",
                drive_conf["device"],
                drive_conf["node_id"],
                drive_conf["baudrate"],
                drive_conf["channel"],
            )
        except Exception as e:
            logger.info(f"Couldn't connect to the drive: {e}")
            continue

        status_callback = partial(logger.info, "Load firmware status: %s")
        progress_callback = partial(logger.info, "Load firmware progress: %s")
        try:
            net.load_firmware(
                servo.target, drive_conf["fw_file"], status_callback, progress_callback
            )
            # Reaching this means that FW was correctly flashed
            break

        except ILFirmwareLoadError as e:
            logger.error(f"CAN boot error: {e}")
            time.sleep(SLEEP_TIME_AFTER_ATTEMP)

        finally:
            try:
                net.disconnect_from_slave(servo)
            except Exception as e:
                logger.error(f"Error when disconnection from drive: {e}")

    logger.info(
        "FW updated. %s, node: %d, baudrate: %d, channel: %d",
        drive_conf["device"],
        drive_conf["node_id"],
        drive_conf["baudrate"],
        drive_conf["channel"],
    )

    logger.info(f"Waiting {SLEEP_TIME_AFTER_BL} seconds for trying to connect")
    time.sleep(SLEEP_TIME_AFTER_BL)

    # Check whether the new FW is present
    detected = False
    ini_time = time.perf_counter()
    while (time.perf_counter() - ini_time) <= TIMEOUT_NEW_FW_DETECT and not detected:
        try:
            net, servo = connect_can(drive_conf)
            # Reaching this point means we are connected
            detected = True
            fw_version = servo.info["firmware_version"]
            logger.info(
                f"New FW detected ({fw_version}) after: {time.perf_counter() - ini_time:.1f} s"
            )
            net.disconnect_from_slave(servo)
        except Exception as e:
            # When cannot connect
            time.sleep(SLEEP_TIME_NEW_FW_DETECT)

    if not detected:
        raise Exception("Could not connect to the drive. FW loading might have failed.")


def load_ecat(drive_conf):
    ifname = None
    for adapter in ifaddr.get_adapters():
        for ip in adapter.ips:
            if ip.is_IPv4 and ip.ip == drive_conf["ip"]:
                ifname = "\\Device\\NPF_{}".format(bytes.decode(adapter.name))
                break
        if ifname is not None:
            break
    net = EthercatNetwork(ifname)
    try:
        net.load_firmware(drive_conf["fw_file"], drive_conf["boot_in_app"], drive_conf["slave"])
    except ILError as e:
        raise Exception(f"Could not load the firmware: {e}")
    logger.info("FW updated. ifname: %s, slave: %d", ifname, drive_conf["slave"])


def ping_check(target_ip, timeout=180):
    # TODO Stop use this function when issue INGM-104 will done
    time.sleep(5)
    initial_time = time.time()
    success_num_pings = 3
    num_pings = 0
    detected = False
    while (time.time() - initial_time) < timeout and not detected:
        aux_ping = ping(target_ip)
        if type(aux_ping) == float:
            num_pings += 1
        if num_pings >= success_num_pings:
            detected = True
        time.sleep(1)
    if not detected:
        logger.error("drive ping not detected", drive=target_ip)
    return detected


def connect_eth(drive_conf):
    net = EthernetNetwork()
    servo = net.connect_to_slave(drive_conf["ip"], drive_conf["dictionary"])
    return net, servo


def boot_mode(net, servo):
    PASSWORD_FORCE_BOOT_COCO = 0x424F4F54
    try:
        servo.write("DRV_BOOT_COCO_FORCE", PASSWORD_FORCE_BOOT_COCO, subnode=0)
        ftp_ready = ping_check(servo.ip_address, timeout=5)
    except ILError as e:
        logger.debug("Could not enter in boot mode.")
        raise e
    finally:
        logger.debug("Disconnecting from drive.")
        net.disconnect_from_slave(servo)
    if not ftp_ready:
        logger.debug("FTP is not ready.")
        raise ILError("FTP is not ready.")
    else:
        logger.debug("FTP is ready.")


def load_eth(drive_conf):
    net, servo = connect_eth(drive_conf)
    logger.info("Drive connected. IP: %s", drive_conf["ip"])
    try:
        boot_mode(net, servo)
    except ILError as e:
        logger.warning(
            f"Drive does not respond ({e}). It may already be in boot mode.", drive=drive_conf["ip"]
        )
    try:
        net.load_firmware(drive_conf["fw_file"], drive_conf["ip"], "Ingenia", "Ingenia")
    except ILError as e:
        raise Exception(f"Could not load the firmware: {e}")
    detected = ping_check(drive_conf["ip"])
    if not detected:
        logger.info("FW not updated. IP: %s", drive_conf["ip"])
        raise Exception("Could not detect the drive.")
    logger.info("FW updated. IP: %s", drive_conf["ip"])


def main(comm, config):
    servo_list = config[comm]
    for index, servo_conf in enumerate(servo_list):
        logger.info("Upload FW comm %s, index: %d", comm, index)
        try:
            if comm == "canopen":
                load_can(servo_conf)
            if comm == "soem":
                load_ecat(servo_conf)
            if comm == "eoe":
                load_eth(servo_conf)
        except ILError as e:
            logger.exception(e)
            logger.error("Error in FW update. comm %s, index: %d", comm, index)


if __name__ == "__main__":
    args = setup_command()
    with open(os.path.join(dirname, "config.json")) as file:
        config_json = json.load(file)
    main(args.comm, config_json)



from typing import TYPE_CHECKING, List


if TYPE_CHECKING:
    from ingeniamotion import MotionController
from ingeniamotion.wizard_tests.feedbacks_tests.feedback_test import Feedbacks
from ingeniamotion.wizard_tests.base_test import BaseTest
from ingeniamotion.enums import SensorType


class DigitalHallTest(Feedbacks):
    HALLS_FILTER_CUTOFF_FREQUENCY = 10
    DIG_HALL_POLE_PAIRS_REGISTER = "FBK_DIGHALL_PAIRPOLES"

    BACKUP_REGISTERS_HALLS: List[str] = [
        "FBK_DIGHALL_POLARITY",
        "FBK_DIGHALL_PAIRPOLES",
        "ERROR_DIGHALL_SEQ_OPTION",
    ]

    FEEDBACK_POLARITY_REGISTER = "FBK_DIGHALL_POLARITY"

    SENSOR_TYPE_FEEDBACK_TEST = SensorType.HALLS

    def __init__(self, mc: "MotionController", servo: str, axis: int) -> None:
        super().__init__(mc, servo, axis)
        self.backup_registers_names.extend(self.BACKUP_REGISTERS_HALLS)

    @BaseTest.stoppable
    def feedback_setting(self) -> None:
        self.halls_extra_settings()
        super().feedback_setting()

    @BaseTest.stoppable
    def halls_extra_settings(self) -> None:
        if self.pair_poles is None:
            raise TypeError("Pair poles has to be an integer")
        self.mc.communication.set_register(
            self.DIG_HALL_POLE_PAIRS_REGISTER, self.pair_poles, servo=self.servo, axis=self.axis
        )

        # Read velocity feedback
        velocity_feedback = self.mc.configuration.get_velocity_feedback(
            servo=self.servo, axis=self.axis
        )
        # Read velocity feedback, if is HALLS set filter to 10 Hz
        # TODO: set filter depending on motors rated velocity by the
        #  following formula: f_halls = w_mechanical * pp * 6
        if velocity_feedback == SensorType.HALLS:
            filter_type_uid = self.VELOCITY_FEEDBACK_FILTER_1_TYPE_REGISTER
            filter_freq_uid = self.VELOCITY_FEEDBACK_FILTER_1_FREQUENCY_REGISTER
            self.suggested_registers[filter_type_uid] = self.LOW_PASS_FILTER
            self.suggested_registers[filter_freq_uid] = self.HALLS_FILTER_CUTOFF_FREQUENCY

            self.logger.info(
                "Setting a velocity low pass filter at 10 Hz as "
                "velocity feedback is set to Halls"
            )
            del self.backup_registers[self.axis][self.VELOCITY_FEEDBACK_FILTER_1_TYPE_REGISTER]
            del self.backup_registers[self.axis][self.VELOCITY_FEEDBACK_FILTER_1_FREQUENCY_REGISTER]

    @BaseTest.stoppable
    def suggest_polarity(self, pol: Feedbacks.Polarity) -> None:
        polarity_uid = self.FEEDBACK_POLARITY_REGISTER
        pair_poles_uid = self.DIG_HALL_POLE_PAIRS_REGISTER
        if self.pair_poles is None:
            raise TypeError("Pair poles has to be set before polarity suggestion.")
        self.suggested_registers[pair_poles_uid] = self.pair_poles
        self.suggested_registers[polarity_uid] = pol



import pytest

from ingeniamotion.enums import SensorCategory, SensorType

COMMUTATION_FEEDBACK_REGISTER = "COMMU_ANGLE_SENSOR"
REFERENCE_FEEDBACK_REGISTER = "COMMU_ANGLE_REF_SENSOR"
VELOCITY_FEEDBACK_REGISTER = "CL_VEL_FBK_SENSOR"
POSITION_FEEDBACK_REGISTER = "CL_POS_FBK_SENSOR"
AUXILIAR_FEEDBACK_REGISTER = "CL_AUX_FBK_SENSOR"
PAIR_POLES_REGISTER = "FBK_DIGHALL_PAIRPOLES"
INCREMENTAL_RESOLUTION_2_REGISTER = "FBK_DIGENC2_RESOLUTION"
INCREMENTAL_RESOLUTION_1_REGISTER = "FBK_DIGENC1_RESOLUTION"
ABS1_1_SINGLE_TURN_REGISTER = "FBK_BISS1_SSI1_POS_ST_BITS"
ABS1_2_SINGLE_TURN_REGISTER = "FBK_BISS2_POS_ST_BITS"
ABS2_1_SINGLE_TURN_REGISTER = "FBK_SSI2_POS_ST_BITS"

SENSOR_TYPE_AND_CATEGORY = [
    (SensorType.ABS1, SensorCategory.ABSOLUTE),
    (SensorType.QEI, SensorCategory.INCREMENTAL),
    (SensorType.HALLS, SensorCategory.ABSOLUTE),
    (SensorType.SSI2, SensorCategory.ABSOLUTE),
    (SensorType.BISSC2, SensorCategory.ABSOLUTE),
    (SensorType.QEI2, SensorCategory.INCREMENTAL),
    (SensorType.INTGEN, SensorCategory.ABSOLUTE),
]


ABSOLUTE_ENCODER_RESOLUTION_TEST_VALUES = [(22, 4194304), (10, 1024), (15, 32768)]

INCREMENTAL_ENCODER_RESOLUTION_TEST_VALUES = [1000, 4000, 6000]


@pytest.fixture
def restore_resolution_registers(motion_controller):
    mc, alias = motion_controller
    registers = [
        PAIR_POLES_REGISTER,
        INCREMENTAL_RESOLUTION_1_REGISTER,
        INCREMENTAL_RESOLUTION_2_REGISTER,
        ABS1_1_SINGLE_TURN_REGISTER,
        ABS1_2_SINGLE_TURN_REGISTER,
        ABS2_1_SINGLE_TURN_REGISTER,
    ]
    registers_values = [
        mc.communication.get_register(register, servo=alias)
        if mc.info.register_exists(register, servo=alias)
        else None
        for register in registers
    ]
    yield
    for register, register_value in zip(registers, registers_values):
        if register_value is not None:
            mc.communication.set_register(register, register_value, servo=alias)


def skip_if_qei2_is_not_available(mc, alias, sensor=SensorType.QEI2):
    if sensor == SensorType.QEI2 and not mc.info.register_exists(
        INCREMENTAL_RESOLUTION_2_REGISTER, servo=alias
    ):
        pytest.skip("Incremental encoder 2 is not available")


@pytest.mark.virtual
@pytest.mark.smoke
@pytest.mark.usefixtures("clean_and_restore_feedbacks")
@pytest.mark.parametrize("sensor", list(SensorType))
def test_get_commutation_feedback(motion_controller, sensor):
    mc, alias = motion_controller
    mc.communication.set_register(COMMUTATION_FEEDBACK_REGISTER, sensor, servo=alias)
    test_feedback = mc.configuration.get_commutation_feedback(servo=alias)
    assert sensor == test_feedback


@pytest.mark.virtual
@pytest.mark.smoke
@pytest.mark.usefixtures("clean_and_restore_feedbacks")
@pytest.mark.parametrize("sensor", list(SensorType))
def test_set_commutation_feedback(motion_controller, sensor):
    mc, alias = motion_controller
    mc.configuration.set_commutation_feedback(sensor, servo=alias)
    register_value = mc.communication.get_register(COMMUTATION_FEEDBACK_REGISTER, servo=alias)
    assert sensor == register_value


@pytest.mark.virtual
@pytest.mark.smoke
@pytest.mark.usefixtures("clean_and_restore_feedbacks")
@pytest.mark.parametrize("sensor, category", SENSOR_TYPE_AND_CATEGORY)
def test_get_commutation_feedback_category(motion_controller, sensor, category):
    mc, alias = motion_controller
    mc.configuration.set_commutation_feedback(sensor, servo=alias)
    test_category = mc.configuration.get_commutation_feedback_category(servo=alias)
    assert test_category == category


@pytest.mark.virtual
@pytest.mark.smoke
@pytest.mark.usefixtures("clean_and_restore_feedbacks")
@pytest.mark.parametrize("sensor", list(SensorType))
def test_get_commutation_feedback_resolution(motion_controller, sensor):
    mc, alias = motion_controller
    skip_if_qei2_is_not_available(mc, alias, sensor=sensor)
    mc.communication.set_register(COMMUTATION_FEEDBACK_REGISTER, sensor, servo=alias)
    if sensor in [SensorType.INTGEN]:
        with pytest.raises(ValueError):
            mc.configuration.get_commutation_feedback_resolution(servo=alias)
    else:
        test_res_1 = mc.configuration.get_commutation_feedback_resolution(servo=alias)
        test_res_2 = mc.configuration.get_feedback_resolution(sensor, servo=alias)
        assert test_res_1 == test_res_2


@pytest.mark.virtual
@pytest.mark.smoke
@pytest.mark.usefixtures("clean_and_restore_feedbacks")
@pytest.mark.parametrize("sensor", list(SensorType))
def test_get_reference_feedback(motion_controller, sensor):
    mc, alias = motion_controller
    mc.communication.set_register(REFERENCE_FEEDBACK_REGISTER, sensor, servo=alias)
    test_feedback = mc.configuration.get_reference_feedback(servo=alias)
    assert sensor == test_feedback


@pytest.mark.virtual
@pytest.mark.smoke
@pytest.mark.usefixtures("clean_and_restore_feedbacks")
@pytest.mark.parametrize("sensor", list(SensorType))
def test_set_reference_feedback(motion_controller, sensor):
    mc, alias = motion_controller
    mc.configuration.set_reference_feedback(sensor, servo=alias)
    register_value = mc.communication.get_register(REFERENCE_FEEDBACK_REGISTER, servo=alias)
    assert sensor == register_value


@pytest.mark.virtual
@pytest.mark.smoke
@pytest.mark.usefixtures("clean_and_restore_feedbacks")
@pytest.mark.parametrize("sensor, category", SENSOR_TYPE_AND_CATEGORY)
def test_get_reference_feedback_category(motion_controller, sensor, category):
    mc, alias = motion_controller
    mc.configuration.set_commutation_feedback(sensor, servo=alias)
    test_category = mc.configuration.get_commutation_feedback_category(servo=alias)
    assert test_category == category


@pytest.mark.virtual
@pytest.mark.smoke
@pytest.mark.usefixtures("clean_and_restore_feedbacks")
@pytest.mark.parametrize("sensor", list(SensorType))
def test_get_reference_feedback_resolution(motion_controller, sensor):
    mc, alias = motion_controller
    skip_if_qei2_is_not_available(mc, alias, sensor=sensor)
    mc.communication.set_register(REFERENCE_FEEDBACK_REGISTER, sensor, servo=alias)
    if sensor in [SensorType.INTGEN]:
        with pytest.raises(ValueError):
            mc.configuration.get_reference_feedback_resolution(servo=alias)
    else:
        test_res_1 = mc.configuration.get_reference_feedback_resolution(servo=alias)
        test_res_2 = mc.configuration.get_feedback_resolution(sensor, servo=alias)
        assert test_res_1 == test_res_2


@pytest.mark.virtual
@pytest.mark.smoke
@pytest.mark.usefixtures("clean_and_restore_feedbacks")
@pytest.mark.parametrize("sensor", list(SensorType))
def test_get_velocity_feedback(motion_controller, sensor):
    mc, alias = motion_controller
    mc.communication.set_register(VELOCITY_FEEDBACK_REGISTER, sensor, servo=alias)
    test_feedback = mc.configuration.get_velocity_feedback(servo=alias)
    assert sensor == test_feedback


@pytest.mark.virtual
@pytest.mark.smoke
@pytest.mark.usefixtures("clean_and_restore_feedbacks")
@pytest.mark.parametrize("sensor", list(SensorType))
def test_set_velocity_feedback(motion_controller, sensor):
    mc, alias = motion_controller
    mc.configuration.set_velocity_feedback(sensor, servo=alias)
    register_value = mc.communication.get_register(VELOCITY_FEEDBACK_REGISTER, servo=alias)
    assert sensor == register_value


@pytest.mark.virtual
@pytest.mark.smoke
@pytest.mark.usefixtures("clean_and_restore_feedbacks")
@pytest.mark.parametrize("sensor, category", SENSOR_TYPE_AND_CATEGORY)
def test_get_velocity_feedback_category(motion_controller, sensor, category):
    mc, alias = motion_controller
    mc.configuration.set_velocity_feedback(sensor, servo=alias)
    test_category = mc.configuration.get_velocity_feedback_category(servo=alias)
    assert test_category == category


@pytest.mark.virtual
@pytest.mark.smoke
@pytest.mark.usefixtures("clean_and_restore_feedbacks")
@pytest.mark.parametrize("sensor", list(SensorType))
def test_get_velocity_feedback_resolution(motion_controller, sensor):
    mc, alias = motion_controller
    skip_if_qei2_is_not_available(mc, alias, sensor=sensor)
    mc.communication.set_register(VELOCITY_FEEDBACK_REGISTER, sensor, servo=alias)
    if sensor in [SensorType.INTGEN]:
        with pytest.raises(ValueError):
            mc.configuration.get_velocity_feedback_resolution(servo=alias)
    else:
        test_res_1 = mc.configuration.get_velocity_feedback_resolution(servo=alias)
        test_res_2 = mc.configuration.get_feedback_resolution(sensor, servo=alias)
        assert test_res_1 == test_res_2


@pytest.mark.virtual
@pytest.mark.smoke
@pytest.mark.usefixtures("clean_and_restore_feedbacks")
@pytest.mark.parametrize("sensor", list(SensorType))
def test_get_position_feedback(motion_controller, sensor):
    mc, alias = motion_controller
    mc.communication.set_register(POSITION_FEEDBACK_REGISTER, sensor, servo=alias)
    test_feedback = mc.configuration.get_position_feedback(servo=alias)
    assert sensor == test_feedback


@pytest.mark.virtual
@pytest.mark.smoke
@pytest.mark.usefixtures("clean_and_restore_feedbacks")
@pytest.mark.parametrize("sensor", list(SensorType))
def test_set_position_feedback(motion_controller, sensor):
    mc, alias = motion_controller
    mc.configuration.set_position_feedback(sensor, servo=alias)
    register_value = mc.communication.get_register(POSITION_FEEDBACK_REGISTER, servo=alias)
    assert sensor == register_value


@pytest.mark.virtual
@pytest.mark.smoke
@pytest.mark.usefixtures("clean_and_restore_feedbacks")
@pytest.mark.parametrize("sensor, category", SENSOR_TYPE_AND_CATEGORY)
def test_get_position_feedback_category(motion_controller, sensor, category):
    mc, alias = motion_controller
    mc.configuration.set_position_feedback(sensor, servo=alias)
    test_category = mc.configuration.get_position_feedback_category(servo=alias)
    assert test_category == category


@pytest.mark.virtual
@pytest.mark.smoke
@pytest.mark.usefixtures("clean_and_restore_feedbacks")
@pytest.mark.parametrize("sensor", list(SensorType))
def test_get_position_feedback_resolution(motion_controller, sensor):
    mc, alias = motion_controller
    skip_if_qei2_is_not_available(mc, alias, sensor=sensor)
    mc.communication.set_register(POSITION_FEEDBACK_REGISTER, sensor, servo=alias)
    if sensor in [SensorType.INTGEN]:
        with pytest.raises(ValueError):
            mc.configuration.get_position_feedback_resolution(servo=alias)
    else:
        test_res_1 = mc.configuration.get_position_feedback_resolution(servo=alias)
        test_res_2 = mc.configuration.get_feedback_resolution(sensor, servo=alias)
        assert test_res_1 == test_res_2


@pytest.mark.virtual
@pytest.mark.smoke
@pytest.mark.usefixtures("clean_and_restore_feedbacks")
@pytest.mark.parametrize(
    "sensor",
    [
        SensorType.ABS1,
        SensorType.QEI,
        SensorType.HALLS,
        SensorType.SSI2,
        SensorType.BISSC2,
        SensorType.QEI2,
    ],
)
def test_get_auxiliar_feedback(motion_controller, sensor):
    mc, alias = motion_controller
    mc.communication.set_register(AUXILIAR_FEEDBACK_REGISTER, sensor, servo=alias)
    test_feedback = mc.configuration.get_auxiliar_feedback(servo=alias)
    assert sensor == test_feedback


@pytest.mark.virtual
@pytest.mark.smoke
@pytest.mark.usefixtures("clean_and_restore_feedbacks")
@pytest.mark.parametrize(
    "sensor",
    [
        SensorType.ABS1,
        SensorType.QEI,
        SensorType.HALLS,
        SensorType.SSI2,
        SensorType.BISSC2,
        SensorType.QEI2,
    ],
)
def test_set_auxiliar_feedback(motion_controller, sensor):
    mc, alias = motion_controller
    mc.configuration.set_auxiliar_feedback(sensor, servo=alias)
    register_value = mc.communication.get_register(AUXILIAR_FEEDBACK_REGISTER, servo=alias)
    assert sensor == register_value


@pytest.mark.virtual
@pytest.mark.smoke
@pytest.mark.usefixtures("clean_and_restore_feedbacks")
@pytest.mark.parametrize(
    "sensor, category",
    [
        (SensorType.ABS1, SensorCategory.ABSOLUTE),
        (SensorType.QEI, SensorCategory.INCREMENTAL),
        (SensorType.HALLS, SensorCategory.ABSOLUTE),
        (SensorType.SSI2, SensorCategory.ABSOLUTE),
        (SensorType.BISSC2, SensorCategory.ABSOLUTE),
        (SensorType.QEI2, SensorCategory.INCREMENTAL),
    ],
)
def test_get_auxiliar_feedback_category(motion_controller, sensor, category):
    mc, alias = motion_controller
    mc.configuration.set_auxiliar_feedback(sensor, servo=alias)
    test_category = mc.configuration.get_auxiliar_feedback_category(servo=alias)
    assert test_category == category


@pytest.mark.virtual
@pytest.mark.smoke
@pytest.mark.usefixtures("clean_and_restore_feedbacks")
@pytest.mark.parametrize(
    "sensor",
    [
        SensorType.ABS1,
        SensorType.QEI,
        SensorType.HALLS,
        SensorType.SSI2,
        SensorType.BISSC2,
        SensorType.QEI2,
    ],
)
def test_get_auxiliar_feedback_resolution(motion_controller, sensor):
    mc, alias = motion_controller
    skip_if_qei2_is_not_available(mc, alias, sensor=sensor)
    mc.communication.set_register(AUXILIAR_FEEDBACK_REGISTER, sensor, servo=alias)
    if sensor in [SensorType.INTGEN]:
        with pytest.raises(ValueError):
            mc.configuration.get_auxiliar_feedback_resolution(servo=alias)
    else:
        test_res_1 = mc.configuration.get_auxiliar_feedback_resolution(servo=alias)
        test_res_2 = mc.configuration.get_feedback_resolution(sensor, servo=alias)
        assert test_res_1 == test_res_2


@pytest.mark.virtual
@pytest.mark.smoke
@pytest.mark.usefixtures("restore_resolution_registers")
@pytest.mark.parametrize("single_turn, resolution", ABSOLUTE_ENCODER_RESOLUTION_TEST_VALUES)
def test_get_absolute_encoder_1_resolution(motion_controller, single_turn, resolution):
    mc, alias = motion_controller
    mc.communication.set_register(ABS1_1_SINGLE_TURN_REGISTER, single_turn, servo=alias)
    test_res = mc.configuration.get_absolute_encoder_1_resolution(servo=alias)
    assert resolution == test_res


@pytest.mark.virtual
@pytest.mark.smoke
@pytest.mark.usefixtures("restore_resolution_registers")
@pytest.mark.parametrize("resolution", INCREMENTAL_ENCODER_RESOLUTION_TEST_VALUES)
def test_get_incremental_encoder_1_resolution(motion_controller, resolution):
    mc, alias = motion_controller
    mc.communication.set_register(INCREMENTAL_RESOLUTION_1_REGISTER, resolution, servo=alias)
    test_res = mc.configuration.get_incremental_encoder_1_resolution(servo=alias)
    assert resolution == test_res


@pytest.mark.virtual
@pytest.mark.smoke
@pytest.mark.usefixtures("restore_resolution_registers")
@pytest.mark.parametrize("pair_poles, resolution", [(1, 6), (10, 60), (4, 24)])
def test_get_digital_halls_resolution(motion_controller, pair_poles, resolution):
    mc, alias = motion_controller
    mc.communication.set_register(PAIR_POLES_REGISTER, pair_poles, servo=alias)
    test_res = mc.configuration.get_digital_halls_resolution(servo=alias)
    assert resolution == test_res


@pytest.mark.virtual
@pytest.mark.smoke
@pytest.mark.usefixtures("restore_resolution_registers")
@pytest.mark.parametrize("single_turn, resolution", ABSOLUTE_ENCODER_RESOLUTION_TEST_VALUES)
def test_get_secondary_ssi_resolution(motion_controller, single_turn, resolution):
    mc, alias = motion_controller
    mc.communication.set_register(ABS2_1_SINGLE_TURN_REGISTER, single_turn, servo=alias)
    test_res = mc.configuration.get_secondary_ssi_resolution(servo=alias)
    assert resolution == test_res


@pytest.mark.virtual
@pytest.mark.smoke
@pytest.mark.usefixtures("restore_resolution_registers")
@pytest.mark.parametrize("single_turn, resolution", ABSOLUTE_ENCODER_RESOLUTION_TEST_VALUES)
def test_get_absolute_encoder_2_resolution(motion_controller, single_turn, resolution):
    mc, alias = motion_controller
    mc.communication.set_register(ABS1_2_SINGLE_TURN_REGISTER, single_turn, servo=alias)
    test_res = mc.configuration.get_absolute_encoder_2_resolution(servo=alias)
    assert resolution == test_res


@pytest.mark.virtual
@pytest.mark.smoke
@pytest.mark.usefixtures("restore_resolution_registers")
@pytest.mark.parametrize("resolution", INCREMENTAL_ENCODER_RESOLUTION_TEST_VALUES)
def test_get_incremental_encoder_2_resolution(motion_controller, resolution):
    mc, alias = motion_controller
    skip_if_qei2_is_not_available(mc, alias)
    mc.communication.set_register(INCREMENTAL_RESOLUTION_2_REGISTER, resolution, servo=alias)
    test_res = mc.configuration.get_incremental_encoder_2_resolution(servo=alias)
    assert resolution == test_res


@pytest.mark.virtual
@pytest.mark.smoke
def test_instance_sensor_type(motion_controller):
    mc, alias = motion_controller
    test_feedback = mc.configuration.get_commutation_feedback(servo=alias)
    assert isinstance(test_feedback, SensorType)


@pytest.mark.virtual
@pytest.mark.smoke
@pytest.mark.parametrize(
    "sensor, register",
    [
        (SensorType.ABS1, "FBK_BISS1_SSI1_POS_POLARITY"),
        (SensorType.QEI, "FBK_DIGENC1_POLARITY"),
        (SensorType.HALLS, "FBK_DIGHALL_POLARITY"),
        (SensorType.SSI2, "FBK_SSI2_POS_POLARITY"),
        (SensorType.BISSC2, "FBK_BISS2_POS_POLARITY"),
        (SensorType.QEI2, "FBK_DIGENC2_POLARITY"),
    ],
)
def test_get_feedback_polarity_register_uid(motion_controller, sensor, register):
    mc, alias = motion_controller
    assert mc.configuration.get_feedback_polarity_register_uid(sensor) == register



from functools import wraps
from typing import TYPE_CHECKING, Callable, Dict, List, Optional, Union

import ingenialogger
import numpy as np
from ingenialink.enums.register import REG_DTYPE, RegCyclicType
from ingenialink.exceptions import ILValueError
from numpy import ndarray
from numpy.typing import ArrayLike, NDArray

from ingeniamotion.enums import MonitoringVersion
from ingeniamotion.exceptions import IMDisturbanceError, IMStatusWordError
from ingeniamotion.metaclass import DEFAULT_AXIS, DEFAULT_SERVO

if TYPE_CHECKING:
    from ingeniamotion.motion_controller import MotionController

# Constants for typing
# TODO: INGM-327
TYPE_MAPPED_REGISTERS_ALL = Dict[str, Union[str, int, List[float]]]
TYPE_MAPPED_REGISTERS_NAME_AXIS = Dict[str, Union[str, int, REG_DTYPE]]
TYPE_MAPPED_REGISTERS_DATA = Dict[str, List[Union[int, float]]]
TYPE_MAPPED_REGISTERS_DATA_NO_KEY = List[Union[int, float]]


def check_disturbance_disabled(
    func: Callable[..., Union[int, float, None]]
) -> Callable[..., Union[int, float, None]]:
    @wraps(func)
    def wrapper(self, *args, **kwargs):  # type: ignore
        disturbance_enabled = self.mc.capture.is_disturbance_enabled(
            servo=self.servo, version=self._version
        )
        if disturbance_enabled:
            raise IMDisturbanceError("Disturbance is enabled")
        return func(self, *args, **kwargs)

    return wrapper


class Disturbance:
    """Class to configure a disturbance in a servo.

    Args:
        mc : MotionController instance.
        servo : servo alias to reference it. ``default`` by default.
    """

    DISTURBANCE_FREQUENCY_DIVIDER_REGISTER = "DIST_FREQ_DIV"
    DISTURBANCE_MAXIMUM_SAMPLE_SIZE_REGISTER = "DIST_MAX_SIZE"
    MONITORING_DISTURBANCE_STATUS_REGISTER = "MON_DIST_STATUS"

    DISTURBANCE_STATUS_ENABLED_BIT = 0x1000  # TODO: Not implemented yet
    MONITORING_STATUS_ENABLED_BIT = 0x1
    REGISTER_MAP_OFFSET = 0x800

    __data_type_size = {
        REG_DTYPE.U8: 1,
        REG_DTYPE.S8: 1,
        REG_DTYPE.U16: 2,
        REG_DTYPE.S16: 2,
        REG_DTYPE.U32: 4,
        REG_DTYPE.S32: 4,
        REG_DTYPE.U64: 8,
        REG_DTYPE.S64: 8,
        REG_DTYPE.FLOAT: 4,
    }

    def __init__(self, mc: "MotionController", servo: str = DEFAULT_SERVO) -> None:
        super().__init__()
        self.mc = mc
        self.servo = servo
        self.mapped_registers: List[TYPE_MAPPED_REGISTERS_NAME_AXIS] = []
        self.sampling_freq: Optional[float] = None
        self._version = mc.capture._check_version(servo)
        self.logger = ingenialogger.get_logger(__name__, drive=mc.servo_name(servo))
        self.max_sample_number = mc.capture.disturbance_max_sample_size(servo)
        if self._version < MonitoringVersion.MONITORING_V3:
            try:
                self.mc.capture.mcb_synchronization(servo=servo)
            except IMStatusWordError:
                self.logger.warning(
                    "MCB could not be synchronized. Motor is enabled.", drive=mc.servo_name(servo)
                )

    @check_disturbance_disabled
    def set_frequency_divider(self, divider: int) -> float:
        """Function to define disturbance frequency with a prescaler. Frequency will be
        ``Position & velocity loop rate frequency / prescaler``,  see
        :func:`ingeniamotion.configuration.Configuration.get_position_and_velocity_loop_rate`
        to know about this frequency. Monitoring/Disturbance must be disabled.

        Args:
            divider : determines disturbance frequency. It must be ``1`` or higher.

        Return:
            Sample period in seconds.

        Raises:
            ValueError: If divider is less than ``1``.
        """
        if divider < 1:
            raise ValueError("divider must be 1 or higher")
        position_velocity_loop_rate = self.mc.configuration.get_position_and_velocity_loop_rate(
            servo=self.servo
        )
        self.sampling_freq = round(position_velocity_loop_rate / divider, 2)
        self.mc.communication.set_register(
            self.DISTURBANCE_FREQUENCY_DIVIDER_REGISTER, divider, servo=self.servo, axis=0
        )
        return 1 / self.sampling_freq

    @check_disturbance_disabled
    def map_registers(
        self,
        registers: Union[TYPE_MAPPED_REGISTERS_NAME_AXIS, List[TYPE_MAPPED_REGISTERS_NAME_AXIS]],
    ) -> float:
        """Map registers to Disturbance. Disturbance must be disabled.

        Args:
            registers : registers to map.
                Each register must be a dict with two keys.

                .. code-block:: python

                    {
                        "name": "CL_POS_SET_POINT_VALUE",  # Register name.
                        "axis": 1  # Register axis.
                        # If it has no axis field, by default axis 1.
                    }

        Returns:
            Max number of samples

        Raises:
            IMDisturbanceError: If the registers is an empty list.
            IMDisturbanceError: If the register is not allowed to be mapped as
                a disturbance register.
            TypeError: If some parameter has a wrong type.
        """
        if len(registers) == 0:
            raise IMDisturbanceError("No registers to be mapped.")
        if not isinstance(registers, List):
            registers = [registers]
        drive = self.mc.servos[self.servo]
        drive.disturbance_remove_all_mapped_registers()
        total_sample_size = 0
        for ch_idx, channel in enumerate(registers):
            subnode = channel.get("axis", DEFAULT_AXIS)
            if not isinstance(subnode, int):
                raise TypeError("Subnode value has to be an integer")
            register = channel["name"]
            if not isinstance(register, str):
                raise TypeError("Register key has to be a string")
            register_obj = self.mc.info.register_info(register, subnode, servo=self.servo)
            dtype = register_obj.dtype
            cyclic = register_obj.cyclic
            if cyclic != RegCyclicType.RX:
                drive.disturbance_remove_all_mapped_registers()
                raise IMDisturbanceError(
                    "{} can not be mapped as a disturbance register".format(register)
                )
            channel["dtype"] = dtype
            drive.disturbance_set_mapped_register(
                ch_idx,
                register_obj.mapped_address,
                subnode,
                dtype.value,
                self.__data_type_size[dtype],
            )
            self.mapped_registers.append(channel)
            total_sample_size += self.__data_type_size[dtype]
        return self.max_sample_number / total_sample_size

    @staticmethod
    def __registers_data_adapter(
        registers_data: Union[
            List[
                Union[
                    int,
                    float,
                    NDArray[np.int32],
                    NDArray[np.float32],
                    TYPE_MAPPED_REGISTERS_DATA_NO_KEY,
                ]
            ],
            NDArray[np.int32],
            NDArray[np.float32],
        ]
    ) -> List[TYPE_MAPPED_REGISTERS_DATA_NO_KEY]:
        if isinstance(registers_data, ndarray):
            registers_data = registers_data.tolist()
            return [registers_data]  # type: ignore [list-item]
        elif isinstance(registers_data, List) and all(
            isinstance(data, List) for data in registers_data
        ):
            return registers_data  # type: ignore [return-value]
        elif isinstance(registers_data, List) and all(
            isinstance(data, (int, float)) for data in registers_data
        ):
            return [registers_data]  # type: ignore [list-item]
        elif isinstance(registers_data, List) and all(
            isinstance(data, ndarray) for data in registers_data
        ):
            for i, x in enumerate(registers_data):
                if isinstance(x, ndarray):
                    registers_data[i] = x.tolist()
            return registers_data  # type: ignore [return-value]
        else:
            raise TypeError(
                "Registers data adapter doesn't have the correct type for its input argument"
            )

    @check_disturbance_disabled
    def write_disturbance_data(
        self,
        registers_data: Union[
            List[
                Union[
                    int,
                    float,
                    NDArray[np.int_],
                    NDArray[np.float_],
                    TYPE_MAPPED_REGISTERS_DATA_NO_KEY,
                ]
            ],
            NDArray[np.int_],
            NDArray[np.float_],
        ],
    ) -> None:
        """Write data in mapped registers. Disturbance must be disabled.

        Args:
            registers_data :
                data to write in disturbance. Registers should have same order
                as in :func:`map_registers`.

        Raises:
            IMDisturbanceError: If there are no mapped registers or the sampling frequency is not
                set yet.
            IMDisturbanceError: If buffer size is not enough for all the
                registers and samples.
        """
        if len(self.mapped_registers) == 0 or self.sampling_freq is None:
            raise IMDisturbanceError("Disturbance is not correctly configured yet")
        adapted_registers_data = self.__registers_data_adapter(registers_data)
        drive = self.mc.servos[self.servo]
        self.__check_buffer_size_is_enough(adapted_registers_data)
        idx_list = list(range(len(adapted_registers_data)))
        dtype_list = [REG_DTYPE(x["dtype"]) for x in self.mapped_registers]
        if self._version >= MonitoringVersion.MONITORING_V3:
            drive.disturbance_remove_data()
        try:
            drive.disturbance_write_data(idx_list, dtype_list, adapted_registers_data)
        except ILValueError as e:
            raise IMDisturbanceError(e)

    def map_registers_and_write_data(
        self, registers: Union[TYPE_MAPPED_REGISTERS_ALL, List[TYPE_MAPPED_REGISTERS_ALL]]
    ) -> None:
        """Map registers to Disturbance and write data. Disturbance must be disabled.

        Args:
            registers : registers to map and write data.
                Each register must be a dict with three keys:

                .. code-block:: python

                    {
                        "name": "CL_POS_SET_POINT_VALUE",  # Register name.
                        "axis": 1,  # Register axis.
                        # If it has no axis field, by default axis 1.
                        "data": [0.0, 0.1, 0.2, ...]  # Data for load in this register
                    }

        Raises:
            IMDisturbanceError: If the register is not allowed to be mapped as a
                disturbance register.
            IMDisturbanceError: If buffer size is not enough for all the
                registers and samples.
        """
        if not isinstance(registers, List):
            registers = [registers]
        registers_keys = []
        registers_data = []
        for channel in registers:
            subnode = channel.get("axis", DEFAULT_AXIS)
            register = channel["name"]
            registers_keys.append({"axis": subnode, "name": register})
            registers_data.append(channel["data"])
        self.map_registers(registers_keys)
        self.write_disturbance_data(registers_data)

    def __check_buffer_size_is_enough(
        self, registers: List[TYPE_MAPPED_REGISTERS_DATA_NO_KEY]
    ) -> None:
        total_buffer_size = 0
        for ch_idx, data in enumerate(registers):
            dtype = self.mapped_registers[ch_idx]["dtype"]
            if not isinstance(dtype, REG_DTYPE):
                continue
            total_buffer_size += self.__data_type_size[dtype] * len(data)
        if total_buffer_size > self.max_sample_number:
            raise IMDisturbanceError(
                "Number of samples is too high. "
                "Demanded size: {} bytes, buffer max size: {} bytes.".format(
                    total_buffer_size, self.max_sample_number
                )
            )
        self.logger.debug(
            "Demanded size: %d bytes, buffer max size: %d bytes.",
            total_buffer_size,
            self.max_sample_number,
        )



import pytest

from ingenialink.ethercat.network import EthercatNetwork
from ingenialink.ethernet.network import EthernetNetwork
from ingenialink.canopen.network import CanopenNetwork, CAN_DEVICE, CAN_BAUDRATE
from ingenialink.dictionary import SubnodeType

from ingeniamotion.enums import REG_DTYPE, REG_ACCESS
from ingeniamotion.information import COMMUNICATION_TYPE
from ingeniamotion.exceptions import IMRegisterNotExist, IMException


@pytest.mark.virtual
@pytest.mark.parametrize(
    "uid, axis",
    [
        ("CL_POS_FBK_VALUE", 1),
        ("CL_VEL_SET_POINT_VALUE", 1),
        ("PROF_POS_OPTION_CODE", 1),
        ("PROF_IP_CLEAR_DATA", 1),
    ],
)
def test_register_info(motion_controller, uid, axis):
    mc, alias = motion_controller
    register = mc.info.register_info(uid, axis, alias)
    assert isinstance(register.dtype, REG_DTYPE)
    assert isinstance(register.access, REG_ACCESS)
    assert isinstance(register.range, tuple)


@pytest.mark.virtual
@pytest.mark.parametrize(
    "uid, axis, dtype",
    [
        ("CL_POS_FBK_VALUE", 1, REG_DTYPE.S32),
        ("CL_VEL_SET_POINT_VALUE", 1, REG_DTYPE.FLOAT),
        ("PROF_POS_OPTION_CODE", 1, REG_DTYPE.U16),
        ("PROF_IP_CLEAR_DATA", 1, REG_DTYPE.U16),
    ],
)
def test_register_type(motion_controller, uid, axis, dtype):
    mc, alias = motion_controller
    register_dtype = mc.info.register_type(uid, axis, alias)
    assert register_dtype == dtype


@pytest.mark.virtual
@pytest.mark.parametrize(
    "uid, axis, access",
    [
        ("CL_POS_FBK_VALUE", 1, REG_ACCESS.RO),
        ("CL_VEL_SET_POINT_VALUE", 1, REG_ACCESS.RW),
        ("PROF_POS_OPTION_CODE", 1, REG_ACCESS.RW),
        ("PROF_IP_CLEAR_DATA", 1, REG_ACCESS.WO),
    ],
)
def test_register_access(motion_controller, uid, axis, access):
    mc, alias = motion_controller
    register_access = mc.info.register_access(uid, axis, alias)
    assert register_access == access


@pytest.mark.virtual
@pytest.mark.parametrize(
    "uid, axis, range",
    [
        ("CL_POS_FBK_VALUE", 1, (-2147483648, 2147483647)),
        ("CL_VEL_SET_POINT_VALUE", 1, (-3.4e38, 3.4e38)),
        ("PROF_POS_OPTION_CODE", 1, (0, 65535)),
        ("PROF_IP_CLEAR_DATA", 1, (0, 65535)),
    ],
)
def test_register_range(motion_controller, uid, axis, range):
    mc, alias = motion_controller
    register_range = mc.info.register_range(uid, axis, alias)
    assert tuple(register_range) == range


@pytest.mark.virtual
@pytest.mark.parametrize(
    "uid, axis, exists",
    [
        ("CL_POS_FBK_VALUE", 1, True),
        ("CL_VEL_SET_POINT_VALUE", 1, True),
        ("PROF_POS_OPTION_CODE", 1, True),
        ("PROF_IP_CLEAR_DATA", 1, True),
        ("DRV_AXIS_NUMBER", 0, True),
        ("WRONG_UID", 1, False),
        ("drv_axis_number", 0, False),
    ],
)
def test_register_exists(motion_controller, uid, axis, exists):
    mc, alias = motion_controller
    register_exists = mc.info.register_exists(uid, axis, alias)
    assert register_exists == exists


@pytest.mark.virtual
def test_get_product_name(motion_controller, mocker):
    expected_product_name = "VIRTUAL-DRIVE"

    mc, alias = motion_controller
    product_name = mc.info.get_product_name(alias)

    assert product_name == expected_product_name


@pytest.mark.virtual
def test_get_ip(motion_controller):
    expected_ip = "127.0.0.1"

    mc, alias = motion_controller
    ip = mc.info.get_ip(alias)

    assert ip == expected_ip


@pytest.mark.virtual
def test_get_name(motion_controller):
    expected_name = "Drive"

    mc, alias = motion_controller
    name = mc.info.get_name(alias)

    assert name == expected_name


@pytest.mark.smoke
@pytest.mark.parametrize(
    "communication, expected_result, args",
    [
        (EthernetNetwork, COMMUNICATION_TYPE.Ethernet, None),
        (EthercatNetwork, COMMUNICATION_TYPE.Ethercat, "fake_interface_name"),
        (CanopenNetwork, COMMUNICATION_TYPE.Canopen, CAN_DEVICE.PCAN),
    ],
)
@pytest.mark.virtual
def test_get_communication_type(mocker, motion_controller, communication, expected_result, args):
    mc, alias = motion_controller
    if communication != EthernetNetwork:
        mocker.patch.object(mc, "_get_network", return_value=communication(args))
    communication_type = mc.info.get_communication_type(alias)
    assert communication_type == expected_result


@pytest.mark.smoke
@pytest.mark.parametrize(
    "communication, expected_result, args",
    [
        (EthernetNetwork, "VIRTUAL-DRIVE - Drive (127.0.0.1)", None),
        (EthercatNetwork, "VIRTUAL-DRIVE - Drive", "fake_interface_name"),
        (CanopenNetwork, "VIRTUAL-DRIVE - Drive", CAN_DEVICE.PCAN),
    ],
)
@pytest.mark.virtual
def test_get_full_name(mocker, motion_controller, communication, expected_result, args):
    mc, alias = motion_controller
    if communication != EthernetNetwork:
        mocker.patch.object(mc, "_get_network", return_value=communication(args))
    full_name = mc.info.get_full_name(alias)
    assert full_name == expected_result


@pytest.mark.virtual
def test_get_subnodes(motion_controller):
    expected_subnodes = 2

    mc, alias = motion_controller
    subnodes = mc.info.get_subnodes(alias)

    assert len(subnodes) == expected_subnodes
    assert subnodes[0] == SubnodeType.COMMUNICATION
    assert subnodes[1] == SubnodeType.MOTION


@pytest.mark.virtual
def test_get_categories(motion_controller):
    expected_number_categories = 19

    mc, alias = motion_controller
    categories = mc.info.get_categories(alias)

    assert len(categories) == expected_number_categories


@pytest.mark.virtual
def test_get_dictionary_file_name(motion_controller):
    expected_dictionary_path = "virtual_drive.xdf"

    mc, alias = motion_controller
    dictionary_file_name = mc.info.get_dictionary_file_name(alias)

    assert dictionary_file_name in expected_dictionary_path


@pytest.mark.virtual
def test_get_encoded_image_from_dictionary(motion_controller):
    expected_type_output = str

    mc, alias = motion_controller
    encoded_image = mc.info.get_encoded_image_from_dictionary(alias)

    assert type(encoded_image) == expected_type_output


@pytest.mark.virtual
def test_register_info_exception(motion_controller):
    mc, alias = motion_controller
    with pytest.raises(IMRegisterNotExist):
        mc.info.register_info("non_existing_uid", 1, alias)


@pytest.mark.virtual
def test_get_product_name_none(motion_controller):
    mc, alias = motion_controller
    drive = mc._get_drive(alias)
    drive.dictionary.part_number = None
    product_name = mc.info.get_product_name(alias)
    assert product_name is None


@pytest.mark.virtual
def test_get_node_id_exception(motion_controller):
    mc, alias = motion_controller
    with pytest.raises(IMException):
        mc.info.get_node_id(alias)


@pytest.mark.virtual
def test_get_ip_exception(mocker, motion_controller):
    mc, alias = motion_controller
    mocker.patch.object(mc, "_get_network", return_value=EthercatNetwork("fake_interface_name"))
    with pytest.raises(IMException):
        mc.info.get_ip(alias)


@pytest.mark.virtual
def test_get_slave_id_exception(motion_controller):
    mc, alias = motion_controller
    with pytest.raises(IMException):
        mc.info.get_slave_id(alias)


@pytest.mark.virtual
def test_get_baudrate_success(motion_controller, mocker):
    mc, _ = motion_controller

    fake_device = CAN_DEVICE.PCAN
    fake_channel = 0
    fake_baudrate = CAN_BAUDRATE.Baudrate_1M
    fake_network = CanopenNetwork(fake_device, fake_channel, fake_baudrate)
    mocker.patch.object(mc, "_get_network", return_value=fake_network)

    test_baudrate = mc.info.get_baudrate()

    assert fake_baudrate == test_baudrate


@pytest.mark.virtual
def test_get_baudrate_failed(motion_controller, mocker):
    mc, _ = motion_controller

    mocker.patch.object(mc, "_get_network", return_value=EthercatNetwork("fake_interface_name"))
    with pytest.raises(IMException) as imexpeption_info:
        _ = mc.info.get_baudrate()

    expected_message_error = "The servo default is not a CANopen device."
    assert expected_message_error == imexpeption_info.value.args[0]



import time

import pytest
import numpy as np

from ingeniamotion.exceptions import IMStatusWordError, IMMonitoringError
from ingeniamotion.enums import (
    OperationMode,
    MonitoringSoCType,
    MonitoringSoCConfig,
    MonitoringVersion,
    MonitoringProcessStage,
)


def __compare_signals(expected_signal, received_signal, fft_tol=0.05):
    fft_received = np.abs(np.fft.fft(received_signal))
    fft_expected = np.abs(np.fft.fft(expected_signal))

    # Normalization
    fft_received = fft_received / np.amax(fft_received)
    fft_expected = fft_expected / np.amax(fft_expected)

    return np.allclose(fft_received, fft_expected, rtol=0, atol=fft_tol)


def test_create_poller(motion_controller):
    registers = [{"name": "CL_CUR_Q_SET_POINT", "axis": 1}]
    sampling_time = 0.125
    mc, alias = motion_controller
    mc.motion.set_current_quadrature(0, servo=alias)
    poller = mc.capture.create_poller(registers, alias, sampling_time, buffer_size=64)
    period = 1
    expected_signal = [0.0] * int(period / sampling_time)
    time.sleep(sampling_time)
    for i in range(7):
        time.sleep(period + sampling_time / 10.0)
        mc.motion.set_current_quadrature(0.2 * (i + 1), servo=alias)
        expected_signal.extend([0.2 * (i + 1)] * int(period / sampling_time))
    time.sleep(3 * period)
    timestamp, test_data, _ = poller.data
    poller.stop()
    assert np.allclose(np.diff(timestamp), sampling_time, rtol=0.5, atol=0)

    received_signal = test_data[0]
    assert len(received_signal) == len(expected_signal)
    assert __compare_signals(expected_signal, received_signal)


def test_create_monitoring_no_trigger(
    skip_if_monitoring_not_available, motion_controller, disable_monitoring_disturbance
):
    registers = [{"name": "CL_POS_REF_VALUE", "axis": 1}]
    mc, alias = motion_controller
    mc.motion.set_operation_mode(OperationMode.VELOCITY, alias)
    max_frequency = mc.configuration.get_position_and_velocity_loop_rate(alias)
    divider = 40
    samples = 2000
    quarter_num_samples = samples // 4
    freq = max_frequency / divider
    total_time = samples / freq
    quarter_total_time = total_time / 4
    mc.motion.move_to_position(0, alias)
    mc.motion.motor_enable(alias)
    monitoring = mc.capture.create_monitoring(registers, divider, total_time, servo=alias)
    mc.capture.enable_monitoring_disturbance(servo=alias)
    # Dummy reading
    monitoring.read_monitoring_data()
    monitoring.rearm_monitoring()
    init = time.time()
    expected_signal = [0] * quarter_num_samples
    for i in range(1, 4):
        while init + i * quarter_total_time > time.time():
            pass
        mc.motion.move_to_position(quarter_num_samples * i, alias)
        expected_signal.extend([i * quarter_num_samples] * quarter_num_samples)
    data = monitoring.read_monitoring_data()
    assert len(data[0]) == len(expected_signal)
    assert __compare_signals(expected_signal, data[0])


@pytest.mark.parametrize(
    "trigger_mode, trigger_config, values_list",
    [
        (
            MonitoringSoCType.TRIGGER_EVENT_EDGE,
            MonitoringSoCConfig.TRIGGER_CONFIG_RISING,
            [0, 0.25, 0.5, 0.75],
        ),
        (
            MonitoringSoCType.TRIGGER_EVENT_EDGE,
            MonitoringSoCConfig.TRIGGER_CONFIG_FALLING,
            [0.75, 0.5, 0.25, 0],
        ),
    ],
)
def test_create_monitoring_edge_trigger(
    skip_if_monitoring_not_available,
    motion_controller,
    trigger_mode,
    trigger_config,
    values_list,
    disable_monitoring_disturbance,
):
    register = {"name": "CL_POS_REF_VALUE", "axis": 1}
    mc, alias = motion_controller
    mc.motion.set_operation_mode(OperationMode.VELOCITY, alias)
    max_frequency = mc.configuration.get_position_and_velocity_loop_rate(alias)
    divider = 40
    samples = 2000
    quarter_num_samples = samples // 4
    trigger_value = ((values_list[1] + values_list[2]) / 2) * samples
    freq = max_frequency / divider
    total_time = samples / freq
    monitoring = mc.capture.create_monitoring(
        [register],
        divider,
        total_time,
        trigger_delay=0,
        trigger_mode=trigger_mode,
        trigger_config=trigger_config,
        trigger_signal=register,
        trigger_value=trigger_value,
        servo=alias,
    )
    mc.motion.move_to_position(int(values_list[0] * samples), alias)
    mc.motion.motor_enable(alias)
    mc.capture.enable_monitoring_disturbance(servo=alias)
    time.sleep(1)
    init = time.time()
    quarter_total_time = total_time / 4
    expected_signal = np.zeros(samples)
    expected_signal[:quarter_num_samples] = int(values_list[0] * samples)
    for i in range(1, 4):
        while init + i * quarter_total_time > time.time():
            pass
        value = int(values_list[i] * samples)
        mc.motion.move_to_position(value, alias)
        expected_signal[i * quarter_num_samples : (i + 1) * quarter_num_samples] = value

    data = monitoring.read_monitoring_data()
    assert len(data[0]) == len(expected_signal)
    assert __compare_signals(expected_signal, data[0])


@pytest.mark.parametrize("trigger_delay_rate", [-1 / 4, 1 / 4])
def test_create_monitoring_trigger_delay(
    skip_if_monitoring_not_available,
    motion_controller,
    trigger_delay_rate,
):
    trigger_mode = MonitoringSoCType.TRIGGER_EVENT_EDGE
    trigger_config = MonitoringSoCConfig.TRIGGER_CONFIG_RISING
    values_list = [0, 0.25, 0.5, 0.75]
    register = {"name": "CL_POS_REF_VALUE", "axis": 1}
    mc, alias = motion_controller
    mc.motion.set_operation_mode(OperationMode.VELOCITY, alias)
    max_frequency = mc.configuration.get_position_and_velocity_loop_rate(alias)
    divider = 40
    samples = 2000
    quarter_num_samples = samples // 4
    trigger_value = ((values_list[1] + values_list[2]) / 2) * samples
    freq = max_frequency / divider
    total_time = samples / freq
    trigger_delay = total_time * trigger_delay_rate
    trigger_delay_samples = int(samples * trigger_delay_rate) + 1
    monitoring = mc.capture.create_monitoring(
        [register],
        divider,
        total_time,
        trigger_delay=trigger_delay,
        trigger_mode=trigger_mode,
        trigger_config=trigger_config,
        trigger_signal=register,
        trigger_value=trigger_value,
        servo=alias,
    )
    mc.motion.move_to_position(int(values_list[0] * samples), alias)
    mc.motion.motor_enable(alias)
    mc.capture.enable_monitoring_disturbance(servo=alias)
    time.sleep(1)
    init = time.time()
    quarter_total_time = total_time / 4
    expected_signal = np.zeros(samples)
    expected_signal[:quarter_num_samples] = int(values_list[0] * samples)
    for i in range(1, 4):
        while init + i * quarter_total_time > time.time():
            pass
        value = int(values_list[i] * samples)
        mc.motion.move_to_position(value, alias)
        expected_signal[i * quarter_num_samples : (i + 1) * quarter_num_samples] = value

    expected_signal = np.roll(expected_signal, -trigger_delay_samples)
    if trigger_delay_samples < 0:
        expected_signal[:-trigger_delay_samples] = 0
    else:
        expected_signal[-trigger_delay_samples:] = value
    data = monitoring.read_monitoring_data()
    assert len(data[0]) == len(expected_signal)
    assert __compare_signals(expected_signal, data[0])


def test_create_disturbance(
    skip_if_monitoring_not_available, motion_controller, disable_monitoring_disturbance
):
    mc, alias = motion_controller
    target_register = "CL_POS_SET_POINT_VALUE"
    max_frequency = mc.configuration.get_position_and_velocity_loop_rate(alias)
    divider = 40
    samples = 2000
    freq = max_frequency / divider
    period = 1 / freq
    data = []
    data_subrange = samples // 4
    for i in range(samples // data_subrange):
        data += [i * data_subrange] * data_subrange
    dist = mc.capture.create_disturbance(target_register, data, divider, servo=alias)
    init_time = time.time()
    mc.capture.enable_monitoring_disturbance(servo=alias)
    read_data = []
    dist_timestamp = np.arange(samples) * period
    read_timestamp = []
    while time.time() < init_time + samples * period:
        read_timestamp.append(time.time() - init_time)
        current_value = mc.communication.get_register(target_register, alias)
        read_data.append(current_value)
        time.sleep(period)

    read_data = np.interp(dist_timestamp, read_timestamp, read_data)
    assert len(data) == len(read_data)
    assert __compare_signals(data, read_data)


@pytest.mark.virtual
@pytest.mark.smoke
def test_mcb_synchronization(mocker, motion_controller):
    mc, alias = motion_controller
    enable_mon = mocker.patch("ingeniamotion.capture.Capture.enable_monitoring")
    disable_mon = mocker.patch("ingeniamotion.capture.Capture.disable_monitoring")
    mc.capture.mcb_synchronization(servo=alias)
    enable_mon.assert_called_once_with(servo=alias)
    disable_mon.assert_called_once_with(servo=alias)


@pytest.mark.smoke
def test_mcb_synchronization_fail(motion_controller):
    mc, alias = motion_controller
    mc.motion.motor_enable(servo=alias)
    with pytest.raises(IMStatusWordError):
        mc.capture.mcb_synchronization(servo=alias)


@pytest.mark.virtual
@pytest.mark.smoke
def test_disturbance_max_sample_size(skip_if_monitoring_not_available, motion_controller):
    mc, alias = motion_controller
    target_register = mc.capture.DISTURBANCE_MAXIMUM_SAMPLE_SIZE_REGISTER
    axis = 0
    max_sample_size = mc.capture.disturbance_max_sample_size(servo=alias)
    drive = mc.servos[alias]
    value = drive.read(target_register, subnode=axis)
    assert max_sample_size == value


@pytest.mark.virtual
@pytest.mark.smoke
def test_monitoring_max_sample_size(skip_if_monitoring_not_available, motion_controller):
    mc, alias = motion_controller
    target_register = mc.capture.MONITORING_MAXIMUM_SAMPLE_SIZE_REGISTER
    axis = 0
    max_sample_size = mc.capture.monitoring_max_sample_size(servo=alias)
    drive = mc.servos[alias]
    value = drive.read(target_register, subnode=axis)
    assert max_sample_size == value


@pytest.mark.smoke
def test_get_frequency(
    skip_if_monitoring_not_available, motion_controller, disable_monitoring_disturbance
):
    mc, alias = motion_controller
    registers = [{"name": "CL_POS_REF_VALUE", "axis": 1}]
    max_frequency = mc.configuration.get_position_and_velocity_loop_rate(alias)
    divider = 40
    samples = 2000
    freq = max_frequency / divider
    total_time = samples / freq
    monitoring = mc.capture.create_monitoring(registers, divider, total_time, servo=alias)
    assert mc.capture.get_frequency(servo=alias) == freq
    new_divider = 2
    monitoring.set_frequency(new_divider)
    assert mc.capture.get_frequency(servo=alias) == max_frequency / new_divider


@pytest.mark.parametrize(
    "name, axis",
    [("CL_CUR_Q_SET_POINT", "1"), (1, 1)],
)
@pytest.mark.virtual
def test_create_poller_exceptions(motion_controller, name, axis):
    sampling_time = 0.0625
    registers = [{"name": name, "axis": axis}]
    mc, alias = motion_controller
    with pytest.raises(TypeError):
        mc.capture.create_poller(registers, alias, sampling_time)


@pytest.mark.virtual
def test_create_empty_monitoring_exception(mocker, motion_controller):
    mc, alias = motion_controller
    mocker.patch.object(mc.capture, "_check_version", return_value=MonitoringVersion.MONITORING_V2)
    with pytest.raises(NotImplementedError):
        mc.capture.create_empty_monitoring(servo=alias)


@pytest.mark.virtual
def test_check_monitoring_version_v3(motion_controller):
    mc, alias = motion_controller
    version = mc.capture._check_version(servo=alias)
    assert version == MonitoringVersion.MONITORING_V3


@pytest.mark.virtual
def test_check_monitoring_version_v2(mocker, motion_controller):
    mc, alias = motion_controller
    mocker.patch.object(mc.capture, "MONITORING_VERSION_REGISTER", return_value="NON_EXISTING_UID")
    version = mc.capture._check_version(servo=alias)
    assert version == MonitoringVersion.MONITORING_V2


@pytest.mark.virtual
def test_check_monitoring_version_v1(mocker, motion_controller):
    mc, alias = motion_controller
    mocker.patch.object(mc.capture, "MONITORING_VERSION_REGISTER", return_value="NON_EXISTING_UID")
    mocker.patch.object(
        mc.capture, "MONITORING_CURRENT_NUMBER_BYTES_REGISTER", return_value="NON_EXISTING_UID"
    )
    version = mc.capture._check_version(servo=alias)
    assert version == MonitoringVersion.MONITORING_V1


@pytest.mark.virtual
def test_check_monitoring_version_not_available(mocker, motion_controller):
    mc, alias = motion_controller
    mocker.patch.object(mc.capture, "MONITORING_VERSION_REGISTER", return_value="NON_EXISTING_UID")
    mocker.patch.object(
        mc.capture, "MONITORING_CURRENT_NUMBER_BYTES_REGISTER", return_value="NON_EXISTING_UID"
    )
    mocker.patch.object(mc.capture, "MONITORING_STATUS_REGISTER", return_value="NON_EXISTING_UID")
    with pytest.raises(NotImplementedError):
        mc.capture._check_version(servo=alias)


@pytest.mark.virtual
def test_enable_monitoring_exception(mocker, motion_controller):
    mc, alias = motion_controller
    monitoring = mc.capture.create_empty_monitoring(alias)
    mocker.patch.object(mc.capture, "is_monitoring_enabled", return_value=False)
    monitoring.map_registers([{"axis": 1, "name": "CL_POS_FBK_VALUE"}])
    with pytest.raises(IMMonitoringError):
        mc.capture.enable_monitoring(servo=alias)


@pytest.mark.virtual
def test_enable_disturbance_exception(mocker, motion_controller):
    mc, alias = motion_controller
    monitoring = mc.capture.create_empty_monitoring(alias)
    mocker.patch.object(mc.capture, "is_disturbance_enabled", return_value=False)
    mocker.patch.object(mc.capture, "is_monitoring_enabled", return_value=False)
    monitoring.map_registers([{"axis": 1, "name": "CL_POS_FBK_VALUE"}])
    with pytest.raises(IMMonitoringError):
        mc.capture.enable_disturbance(servo=alias)


@pytest.mark.parametrize(
    "function",
    ["get_monitoring_disturbance_status", "get_monitoring_status", "get_disturbance_status"],
)
@pytest.mark.virtual
def test_get_monitoring_disturbance_status_exception(mocker, motion_controller, function):
    mc, alias = motion_controller
    mocker.patch.object(mc.communication, "get_register", return_value="invalid_value")
    with pytest.raises(TypeError):
        getattr(mc.capture, function)(servo=alias)


@pytest.mark.parametrize(
    "monitor_status, expected_stage",
    [
        (0x00, MonitoringProcessStage.INIT_STAGE),
        (0x02, MonitoringProcessStage.FILLING_DELAY_DATA),
        (0x04, MonitoringProcessStage.WAITING_FOR_TRIGGER),
        (0x06, MonitoringProcessStage.DATA_ACQUISITION),
        (0x08, MonitoringProcessStage.END_STAGE),
    ],
)
@pytest.mark.virtual
def test_get_monitoring_process_stage_v3(mocker, motion_controller, monitor_status, expected_stage):
    mc, alias = motion_controller
    mocker.patch.object(mc.capture, "get_monitoring_status", return_value=monitor_status)
    assert mc.capture.get_monitoring_process_stage(servo=alias) == expected_stage


@pytest.mark.parametrize(
    "monitoring_status, expected_stage",
    [
        (0x00, MonitoringProcessStage.INIT_STAGE),
        (0x02, MonitoringProcessStage.FILLING_DELAY_DATA),
        (0x04, MonitoringProcessStage.WAITING_FOR_TRIGGER),
        (0x06, MonitoringProcessStage.DATA_ACQUISITION),
    ],
)
@pytest.mark.virtual
def test_get_monitoring_process_stage_v1_v2(
    mocker, motion_controller, monitoring_status, expected_stage
):
    mc, alias = motion_controller
    mocker.patch.object(mc.capture, "get_monitoring_status", return_value=monitoring_status)
    assert (
        mc.capture.get_monitoring_process_stage(
            servo=alias, version=MonitoringVersion.MONITORING_V2
        )
        == expected_stage
    )


@pytest.mark.parametrize(
    "monitoring_status, monitoring_version",
    [
        (0x800, MonitoringVersion.MONITORING_V1),
        (0x800, MonitoringVersion.MONITORING_V2),
        (0x10, MonitoringVersion.MONITORING_V3),
    ],
)
@pytest.mark.virtual
def test_is_frame_available(mocker, motion_controller, monitoring_status, monitoring_version):
    mc, alias = motion_controller
    mocker.patch.object(mc.capture, "get_monitoring_status", return_value=monitoring_status)
    assert mc.capture.is_frame_available(servo=alias, version=monitoring_version)


@pytest.mark.parametrize(
    "function",
    ["disturbance_max_sample_size", "monitoring_max_sample_size"],
)
@pytest.mark.virtual
def test_monitoring_disturbance_max_sample_size_exception(mocker, motion_controller, function):
    mc, alias = motion_controller
    mocker.patch.object(mc.communication, "get_register", return_value="invalid_value")
    with pytest.raises(TypeError):
        getattr(mc.capture, function)(servo=alias)


@pytest.mark.virtual
def test_get_frequency_exception(mocker, motion_controller):
    mc, alias = motion_controller
    mocker.patch.object(mc.communication, "get_register", return_value="invalid_value")
    with pytest.raises(TypeError):
        mc.capture.get_frequency(servo=alias)



import re

import setuptools

with open("docs/what_is_ingeniamotion.rst", "r", encoding="utf-8") as fh:
    long_description = fh.read()

with open("ingeniamotion/__init__.py") as f:
    __version = re.search(r"__version__\s+=\s+\"(.*)\"", f.read()).group(1)


def get_docs_url():
    return "https://distext.ingeniamc.com/doc/ingeniamotion/{}".format(__version)


setuptools.setup(
    name="ingeniamotion",
    version=__version,
    packages=setuptools.find_packages(),
    include_package_data=True,
    package_data={"ingeniamotion": ["py.typed"]},
    author="Ingenia Motion Control",
    author_email="support@ingeniamc.com",
    description="Motion library for Ingenia servo drives",
    long_description=long_description,
    long_description_content_type="text/x-rst",
    url="https://www.ingeniamc.com",
    project_urls={
        "Documentation": get_docs_url(),
        "Source": "https://github.com/ingeniamc/ingeniamotion",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        "ingenialink>=7.3.2, < 8.0.0",
        "ingenialogger>=0.2.1",
        "ifaddr==0.1.7",
    ],
    python_requires=">=3.9",
)
